# coding: utf-8
"""

"""

import datetime
import http
import json
import secrets
import typing

import flask
from flask_babel import refresh, _, lazy_gettext
import flask_login
from fido2.webauthn import PublicKeyCredentialUserEntity, ResidentKeyRequirement, UserVerificationRequirement
import pytz

from ... import db

from .. import frontend
from ..authentication_forms import ChangeUserForm, AuthenticationForm, AuthenticationMethodForm
from ..users_forms import RequestPasswordResetForm, PasswordForm, AuthenticationPasswordForm
from ..permission_forms import handle_permission_forms, set_up_permissions_forms
from .forms import NotificationModeForm, OtherSettingsForm, CreateAPITokenForm, ManageTwoFactorAuthenticationMethodForm, \
    AddWebhookForm, RemoveWebhookForm
from ..utils import get_groups_form_data

from ... import logic
from ...logic import user_log, errors
from ...logic.authentication import add_authentication_method, remove_authentication_method, change_password_in_authentication_method, add_api_token, get_two_factor_authentication_methods, activate_two_factor_authentication_method, deactivate_two_factor_authentication_method, delete_two_factor_authentication_method, get_all_fido2_passkey_credentials, add_fido2_passkey, get_webauthn_server, get_api_tokens, get_authentication_methods, get_authentication_method, confirm_authentication_method_by_email, ALL_AUTHENTICATION_TYPES
from ...logic.users import get_user, get_users
from ...logic.utils import send_email_confirmation_email, send_recovery_email
from ...logic.security_tokens import verify_token
from ...logic.default_permissions import default_permissions, get_default_permissions_for_users, get_default_permissions_for_groups, get_default_permissions_for_projects, get_default_permissions_for_all_users, get_default_permissions_for_anonymous_users
from ...logic.projects import get_project
from ...logic.groups import get_group
from ...logic.notifications import get_notification_modes, set_notification_mode_for_type
from ...logic.settings import get_user_settings, set_user_settings
from ...logic.locale import SUPPORTED_LOCALES
from ...logic.webhooks import get_webhooks, create_webhook, remove_webhook
from ...logic.instruments import get_user_instruments
from ...models import AuthenticationType, Permissions, NotificationType, NotificationMode, BackgroundTaskStatus
from ...models.webhooks import WebhookType
from ...utils import FlaskResponseT


@frontend.route('/users/me/preferences', methods=['GET', 'POST'])
def user_me_preferences() -> FlaskResponseT:
    if flask_login.current_user.is_authenticated:
        return flask.redirect(flask.url_for('.user_preferences', user_id=flask_login.current_user.id))
    return email_for_resetting_password()


@frontend.route('/users/<int:user_id>/preferences', methods=['GET', 'POST'])
def user_preferences(user_id: int) -> FlaskResponseT:
    token = flask.request.args.get('token')
    if token:
        data = verify_token(token, salt='password', secret_key=flask.current_app.config['SECRET_KEY'])
        if data is not None:
            return reset_password()
        else:
            # es ist egal, ob eingeloggt oder nicht
            return confirm_email()
    elif flask_login.current_user.is_authenticated:
        if user_id != flask_login.current_user.id:
            return flask.abort(403)
        else:
            if not flask_login.login_fresh():
                # ensure only fresh sessions can edit preferences including passwords and api tokens
                return flask.redirect(flask.url_for('.refresh_sign_in', next=flask.url_for('.user_preferences', user_id=flask_login.current_user.id)))
            # user is logged in and can edit their preferences
            return change_preferences()
    else:
        return typing.cast(flask_login.LoginManager, flask.current_app.login_manager).unauthorized()  # type: ignore[attr-defined, no-any-return]


def _handle_account_information_forms(
        template_kwargs: typing.Dict[str, typing.Any]
) -> typing.Optional[FlaskResponseT]:
    change_user_form = ChangeUserForm()

    template_kwargs.update(
        change_user_form=change_user_form,
        EXTRA_USER_FIELDS=flask.current_app.config['EXTRA_USER_FIELDS'],
    )
    if 'change' not in flask.request.form:
        if change_user_form.name.data is None or change_user_form.name.data == "":
            change_user_form.name.data = flask_login.current_user.name
        if change_user_form.email.data is None or change_user_form.email.data == "":
            change_user_form.email.data = flask_login.current_user.email
        if change_user_form.orcid.data is None or change_user_form.orcid.data == "":
            change_user_form.orcid.data = flask_login.current_user.orcid
        if change_user_form.affiliation.data is None or change_user_form.affiliation.data == "":
            change_user_form.affiliation.data = flask_login.current_user.affiliation
        if change_user_form.role.data is None or change_user_form.role.data == "":
            change_user_form.role.data = flask_login.current_user.role
    if 'change' in flask.request.form and flask.request.form['change'] == 'Change':
        if change_user_form.validate_on_submit():
            if change_user_form.name.data != flask_login.current_user.name:
                logic.users.update_user(
                    flask_login.current_user.id,
                    updating_user_id=flask_login.current_user.id,
                    name=str(change_user_form.name.data)
                )
                user_log.edit_user_preferences(user_id=flask_login.current_user.id)
                flask.flash(_("Successfully updated your user name."), 'success')
            if change_user_form.email.data != flask_login.current_user.email:
                # send confirm link
                mail_send_status = send_email_confirmation_email(
                    email=change_user_form.email.data,
                    user_id=flask_login.current_user.id,
                    salt='edit_profile'
                )[0]
                if mail_send_status == BackgroundTaskStatus.FAILED:
                    flask.flash(_("Sending an email failed. Please try again later or contact an administrator."), 'error')
                else:
                    flask.flash(_("Please see your email to confirm this change."), 'success')
            if change_user_form.orcid.data != flask_login.current_user.orcid or change_user_form.affiliation.data != flask_login.current_user.affiliation or change_user_form.role.data != flask_login.current_user.role:
                if change_user_form.orcid.data and change_user_form.orcid.data.strip():
                    orcid = change_user_form.orcid.data.strip()
                else:
                    orcid = None
                if change_user_form.affiliation.data and change_user_form.affiliation.data.strip():
                    affiliation = change_user_form.affiliation.data.strip()
                else:
                    affiliation = None
                if change_user_form.role.data and change_user_form.role.data.strip():
                    role = change_user_form.role.data.strip()
                else:
                    role = None
                extra_fields = {}
                for extra_field_id in flask.current_app.config['EXTRA_USER_FIELDS']:
                    extra_field_value = flask.request.form.get('extra_field_' + str(extra_field_id))
                    if extra_field_value:
                        extra_fields[extra_field_id] = extra_field_value
                change_orcid = (flask_login.current_user.orcid != orcid and (orcid is not None or flask_login.current_user.orcid is not None))
                change_affiliation = (flask_login.current_user.affiliation != affiliation and (affiliation is not None or flask_login.current_user.affiliation is not None))
                change_role = (flask_login.current_user.role != role and (role is not None or flask_login.current_user.role is not None))
                change_extra_fields = flask_login.current_user.extra_fields != extra_fields
                if change_orcid or change_affiliation or change_role or change_extra_fields:
                    logic.users.update_user(
                        flask_login.current_user.id,
                        updating_user_id=flask_login.current_user.id,
                        orcid=orcid,
                        affiliation=affiliation,
                        role=role,
                        extra_fields=extra_fields
                    )
                    user_log.edit_user_preferences(user_id=flask_login.current_user.id)
                    flask.flash(_("Successfully updated your user information."), 'success')

            return flask.redirect(flask.url_for('frontend.user_me_preferences'))
    return None


def _handle_authentication_methods_forms(
        template_kwargs: typing.Dict[str, typing.Any]
) -> typing.Optional[FlaskResponseT]:
    authentication_methods = get_authentication_methods(
        user_id=flask_login.current_user.id,
        authentication_types=ALL_AUTHENTICATION_TYPES - {
            AuthenticationType.API_TOKEN,
            AuthenticationType.API_ACCESS_TOKEN
        } - ({AuthenticationType.FEDERATED_LOGIN} if not flask.current_app.config['ENABLE_FEDERATED_LOGIN'] else set())
    )
    authentication_method_ids = [authentication_method.id for authentication_method in authentication_methods]

    authentication_form = AuthenticationForm()
    if flask.current_app.config["ENABLE_FIDO2_PASSKEY_AUTHENTICATION"]:
        authentication_form.authentication_method.choices.append(('FIDO2_PASSKEY', 'FIDO2 Passkey'))
    authentication_method_form = AuthenticationMethodForm()
    authentication_password_form = AuthenticationPasswordForm()

    server = get_webauthn_server()
    all_credentials = get_all_fido2_passkey_credentials()
    options, state = server.register_begin(
        PublicKeyCredentialUserEntity(
            id=f"signin-{flask_login.current_user.id}".encode('utf-8'),
            name=f"{flask_login.current_user.name} ({flask_login.current_user.email})",
            display_name=flask_login.current_user.name,
        ),
        credentials=all_credentials,
        resident_key_requirement=ResidentKeyRequirement.PREFERRED,
        user_verification=UserVerificationRequirement.REQUIRED
    )
    if not authentication_form.validate_on_submit():
        flask.session["webauthn_enroll_state"] = state

    template_kwargs.update(
        authentication_password_form=authentication_password_form,
        authentication_method_form=authentication_method_form,
        authentication_form=authentication_form,
        confirmed_authentication_methods=len([
            authentication_method
            for authentication_method in authentication_methods
            if authentication_method.confirmed and authentication_method.type != AuthenticationType.FIDO2_PASSKEY
        ]),
        authentications=authentication_methods,
        api_access_tokens=get_authentication_methods(
            user_id=flask_login.current_user.id,
            authentication_types={AuthenticationType.API_ACCESS_TOKEN}
        ),
        options=dict(options),
    )

    if 'edit' in flask.request.form and flask.request.form['edit'] == 'Edit':
        if authentication_password_form.validate_on_submit() and authentication_password_form.id.data in authentication_method_ids:
            authentication_method_id = authentication_password_form.id.data
            try:
                password_was_changed = change_password_in_authentication_method(authentication_method_id, authentication_password_form.password.data)
            except Exception as e:
                password_was_changed = False
                template_kwargs.update(
                    error=str(e),
                )
            if not password_was_changed:
                flask.flash(_("Failed to change password."), 'error')
                return None
            flask.flash(_("Successfully updated your password."), 'success')
            user_log.edit_user_preferences(user_id=flask_login.current_user.id)
            return flask.redirect(flask.url_for('frontend.user_me_preferences'))
        else:
            flask.flash(_("Failed to change password."), 'error')
    if 'remove' in flask.request.form and flask.request.form['remove'] == 'Remove':
        authentication_method_id = authentication_method_form.id.data
        if authentication_method_form.validate_on_submit():
            try:
                remove_authentication_method(authentication_method_id)
            except Exception as e:
                flask.flash(_("Failed to remove the authentication method."), 'error')
                template_kwargs.update(
                    error=str(e),
                )
                return None
            else:
                flask.flash(_("Successfully removed the authentication method."), 'success')
                user_log.edit_user_preferences(user_id=flask_login.current_user.id)
                return flask.redirect(flask.url_for('frontend.user_me_preferences'))
    if 'add' in flask.request.form and flask.request.form['add'] == 'Add':
        if logic.oidc.is_oidc_only_auth_method():
            # The form isn't shown to the user, so ignore attempts.
            return None

        if authentication_form.validate_on_submit():
            # check, if login already exists
            all_authentication_methods = {
                'E': AuthenticationType.EMAIL,
                'L': AuthenticationType.LDAP,
                'O': AuthenticationType.OTHER
            }
            if flask.current_app.config['ENABLE_FIDO2_PASSKEY_AUTHENTICATION']:
                all_authentication_methods['FIDO2_PASSKEY'] = AuthenticationType.FIDO2_PASSKEY
            if authentication_form.authentication_method.data not in all_authentication_methods:
                return flask.abort(400)

            authentication_method = all_authentication_methods[authentication_form.authentication_method.data]
            # check, if additional authentication is correct
            try:
                if authentication_method == AuthenticationType.FIDO2_PASSKEY:
                    auth_data = server.register_complete(
                        flask.session["webauthn_enroll_state"],
                        json.loads(authentication_form.fido2_passkey_credentials.data)
                    )
                    assert auth_data.credential_data is not None
                    add_fido2_passkey(
                        user_id=flask_login.current_user.id,
                        credential_data=auth_data.credential_data,
                        description=authentication_form.description.data
                    )
                    del flask.session["webauthn_enroll_state"]
                else:
                    add_authentication_method(flask_login.current_user.id, authentication_form.login.data, authentication_form.password.data, authentication_method)
                flask.flash(_("Successfully added the authentication method."), 'success')
                return flask.redirect(flask.url_for('.user_me_preferences'))
            except Exception as e:
                flask.flash(_("Failed to add an authentication method."), 'error')
                template_kwargs.update(
                    error_add=str(e)
                )
                return None
        else:
            flask.flash(_("Failed to add an authentication method."), 'error')
    return None


def _handle_create_api_token_form(
        template_kwargs: typing.Dict[str, typing.Any]
) -> typing.Optional[FlaskResponseT]:
    create_api_token_form = CreateAPITokenForm()

    template_kwargs.update(
        api_tokens=get_api_tokens(flask_login.current_user.id),
        create_api_token_form=create_api_token_form,
        created_api_token=None
    )
    if 'create_api_token' in flask.request.form and create_api_token_form.validate_on_submit():
        created_api_token = secrets.token_hex(32)
        description = create_api_token_form.description.data
        try:
            add_api_token(flask_login.current_user.id, created_api_token, description)
            template_kwargs.update(
                api_tokens=get_api_tokens(flask_login.current_user.id),
                created_api_token=created_api_token,
            )
        except Exception as e:
            flask.flash(_("Failed to add an API token."), 'error')
            template_kwargs.update(
                error_add=str(e)
            )
    return None


def _handle_two_factor_authentication_forms(
        template_kwargs: typing.Dict[str, typing.Any]
) -> typing.Optional[FlaskResponseT]:
    two_factor_authentication_methods = get_two_factor_authentication_methods(flask_login.current_user.id)
    manage_two_factor_authentication_method_form = ManageTwoFactorAuthenticationMethodForm()
    template_kwargs.update(
        two_factor_authentication_methods=two_factor_authentication_methods,
        manage_two_factor_authentication_method_form=manage_two_factor_authentication_method_form,
        has_active_method=any(method.active for method in two_factor_authentication_methods),
    )

    if manage_two_factor_authentication_method_form.validate_on_submit():
        method_id = manage_two_factor_authentication_method_form.method_id.data
        method = {
            method.id: method
            for method in two_factor_authentication_methods
        }.get(manage_two_factor_authentication_method_form.method_id.data)
        if method is not None:
            if manage_two_factor_authentication_method_form.action.data == 'delete':
                if method.active:
                    flask.flash(_('You cannot delete an active two-factor authentication method.'), 'error')
                    return flask.redirect(flask.url_for('.user_me_preferences'))
                delete_two_factor_authentication_method(method_id)
                flask.flash(_('The two-factor authentication method has been deleted.'), 'success')
                return flask.redirect(flask.url_for('.user_me_preferences'))
            if manage_two_factor_authentication_method_form.action.data == 'enable':
                flask.session['confirm_data'] = {
                    'reason': 'activate_two_factor_authentication_method',
                    'user_id': method.user_id,
                    'expiration_datetime': (datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(minutes=5)).strftime('%Y-%m-%d %H:%M:%S')
                }
                if method.data.get('type') == 'totp':
                    return flask.redirect(flask.url_for('.confirm_totp_two_factor_authentication', method_id=method.id))
                elif method.data.get('type') == 'fido2_passkey':
                    return flask.redirect(flask.url_for('.confirm_fido2_passkey_two_factor_authentication', method_id=method.id))
                else:
                    del flask.session['confirm_data']
                    activate_two_factor_authentication_method(method_id)
                    flask.flash(_('The two-factor authentication method has been enabled.'), 'success')
                    return flask.redirect(flask.url_for('.user_me_preferences'))
            if manage_two_factor_authentication_method_form.action.data == 'disable':
                flask.session['confirm_data'] = {
                    'reason': 'deactivate_two_factor_authentication_method',
                    'user_id': method.user_id,
                    'method_to_deactivate_id': method.id,
                    'expiration_datetime': (datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(minutes=5)).strftime('%Y-%m-%d %H:%M:%S')
                }
                usable_active_two_factor_authentication_methods = [
                    other_method
                    for other_method in two_factor_authentication_methods
                    if other_method.active and other_method.data.get('type') in ('totp', 'fido2_passkey',)
                ]
                if len(usable_active_two_factor_authentication_methods) < 2:
                    if usable_active_two_factor_authentication_methods:
                        two_factor_authentication_method = usable_active_two_factor_authentication_methods[0]
                        if two_factor_authentication_method.data.get('type') == 'totp':
                            return flask.redirect(flask.url_for('.confirm_totp_two_factor_authentication', method_id=two_factor_authentication_method.id))
                        if two_factor_authentication_method.data.get('type') == 'fido2_passkey':
                            return flask.redirect(flask.url_for('.confirm_fido2_passkey_two_factor_authentication', method_id=two_factor_authentication_method.id))
                    del flask.session['confirm_data']
                    deactivate_two_factor_authentication_method(method_id)
                    flask.flash(_('The two-factor authentication method has been disabled.'), 'success')
                    return flask.redirect(flask.url_for('.user_me_preferences'))
                return flask.render_template(
                    'two_factor_authentication/pick.html',
                    methods=usable_active_two_factor_authentication_methods,
                    method_to_deactivate=method
                )
        flask.flash(_('Something went wrong, please try again.'), 'error')
    return None


def _handle_webhook_forms(
        template_kwargs: typing.Dict[str, typing.Any]
) -> typing.Optional[FlaskResponseT]:
    may_use_webhooks = flask_login.current_user.is_admin or flask.current_app.config['ENABLE_WEBHOOKS_FOR_USERS']
    add_webhook_form = AddWebhookForm()
    if add_webhook_form.address.data is None:
        add_webhook_form.address.data = ''
    if add_webhook_form.name.data is None:
        add_webhook_form.name.data = ''
    remove_webhook_form = RemoveWebhookForm()

    template_kwargs.update(
        may_use_webhooks=may_use_webhooks,
        webhooks=get_webhooks(user_id=flask_login.current_user.get_id()),
        show_add_form=False,
        webhook_secret=None,
        add_webhook_form=add_webhook_form,
        remove_webhook_form=remove_webhook_form,
    )

    if 'remove_webhook' in flask.request.form and may_use_webhooks:
        if remove_webhook_form.validate_on_submit():
            try:
                webhook_id = remove_webhook_form.id.data
                remove_webhook(webhook_id)
                flask.flash(_('Successfully removed the webhook.'), 'success')
            except Exception:
                flask.flash(_('Failed to remove the webhook.'), 'error')
            return flask.redirect(flask.url_for('.user_preferences', user_id=flask_login.current_user.id))

    if 'add_webhook' in flask.request.form:
        if not may_use_webhooks:
            flask.flash(_('You are not allowed to create Webhooks.'), 'error')
            return None
        if add_webhook_form.validate_on_submit():
            try:
                name = add_webhook_form.name.data
                address = add_webhook_form.address.data
                if name == '':
                    name = None
                if address == '':
                    address = None
                new_webhook = create_webhook(type=WebhookType.OBJECT_LOG, user_id=flask_login.current_user.get_id(), target_url=address, name=name)
            except errors.WebhookAlreadyExistsError:
                add_webhook_form.address.errors.append(_('A webhook of this type with this target address already exists', service_name=flask.current_app.config['SERVICE_NAME']))
            except errors.InsecureComponentAddressError:
                add_webhook_form.address.errors.append(_('Only secure communication via https is allowed'))
            except errors.InvalidComponentAddressError:
                add_webhook_form.address.errors.append(_('This webhook address is invalid'))
            except Exception:
                add_webhook_form.name.errors.append(_('Failed to create webhook'))
            else:
                flask.flash(_('The webhook has been added successfully'), 'success')
                template_kwargs.update(
                    webhooks=get_webhooks(user_id=flask_login.current_user.get_id()),
                    webhook_secret=new_webhook.secret,
                )
                return None
        template_kwargs.update(
            show_add_form=True
        )
    return None


def _handle_notification_forms(
        template_kwargs: typing.Dict[str, typing.Any]
) -> typing.Optional[FlaskResponseT]:
    notification_mode_form = NotificationModeForm()

    template_kwargs.update(
        notification_modes=get_notification_modes(flask_login.current_user.id),
        notification_mode_form=notification_mode_form,
        NotificationMode=NotificationMode,
        NotificationType=NotificationType,
        is_instrument_responsible_user=bool(get_user_instruments(flask_login.current_user.id, exclude_hidden=True))
    )

    if 'edit_notification_settings' in flask.request.form and notification_mode_form.validate_on_submit():
        for notification_type in NotificationType:
            if 'notification_mode_for_type_' + notification_type.name.lower() in flask.request.form:
                notification_mode_text = flask.request.form.get('notification_mode_for_type_' + notification_type.name.lower())
                for notification_mode in [NotificationMode.IGNORE, NotificationMode.WEBAPP, NotificationMode.EMAIL]:
                    if notification_mode_text == notification_mode.name.lower():
                        set_notification_mode_for_type(notification_type, flask_login.current_user.id, notification_mode)
                        break
        flask.flash(_("Successfully updated your notification settings."), 'success')
        return flask.redirect(flask.url_for('.user_preferences', user_id=flask_login.current_user.id))
    return None


def _handle_default_permissions_forms(
        template_kwargs: typing.Dict[str, typing.Any]
) -> typing.Optional[FlaskResponseT]:
    user_permissions = get_default_permissions_for_users(creator_id=flask_login.current_user.id)
    group_permissions = get_default_permissions_for_groups(creator_id=flask_login.current_user.id)
    project_permissions = get_default_permissions_for_projects(creator_id=flask_login.current_user.id)
    all_user_permissions = get_default_permissions_for_all_users(creator_id=flask_login.current_user.id)
    anonymous_user_permissions = get_default_permissions_for_anonymous_users(creator_id=flask_login.current_user.id)

    (
        add_user_permissions_form,
        add_group_permissions_form,
        add_project_permissions_form,
        default_permissions_form
    ) = set_up_permissions_forms(
        resource_permissions=logic.default_permissions.default_permissions,
        resource_id=flask_login.current_user.id,
        existing_all_user_permissions=all_user_permissions,
        existing_anonymous_user_permissions=anonymous_user_permissions,
        existing_user_permissions=user_permissions,
        existing_group_permissions=group_permissions,
        existing_project_permissions=project_permissions
    )

    users_without_permissions = get_users(exclude_hidden=not flask_login.current_user.is_admin or not flask_login.current_user.settings['SHOW_HIDDEN_USERS_AS_ADMIN'], exclude_fed=True, exclude_eln_import=True)
    users_without_permissions = [user for user in users_without_permissions if user.id not in user_permissions]
    users_without_permissions.sort(key=lambda user: user.id)

    show_groups_form, groups_treepicker_info = get_groups_form_data(
        basic_group_filter=lambda group: group.id not in group_permissions
    )

    show_projects_form, projects_treepicker_info = get_groups_form_data(
        project_group_filter=lambda group: group.id not in project_permissions
    )
    template_kwargs.update(
        Permissions=Permissions,
        user_permissions=user_permissions,
        group_permissions=group_permissions,
        project_permissions=project_permissions,
        all_user_permissions=all_user_permissions,
        default_permissions_form=default_permissions_form,
        add_user_permissions_form=add_user_permissions_form,
        add_group_permissions_form=add_group_permissions_form,
        add_project_permissions_form=add_project_permissions_form,
        users=users_without_permissions,
        show_groups_form=show_groups_form,
        groups_treepicker_info=groups_treepicker_info,
        show_projects_form=show_projects_form,
        projects_treepicker_info=projects_treepicker_info,
        get_user=get_user,
        get_group=get_group,
        get_project=get_project,
    )
    if handle_permission_forms(
        default_permissions,
        flask_login.current_user.id,
        add_user_permissions_form,
        add_group_permissions_form,
        add_project_permissions_form,
        default_permissions_form
    ):
        flask.flash(_("Successfully updated default permissions."), 'success')
        return flask.redirect(flask.url_for('.user_preferences', user_id=flask_login.current_user.id))
    return None


def _handle_other_settings_forms(
        template_kwargs: typing.Dict[str, typing.Any]
) -> typing.Optional[FlaskResponseT]:
    other_settings_form = OtherSettingsForm()
    all_timezones = list(pytz.all_timezones)

    template_kwargs.update(
        user_settings=get_user_settings(flask_login.current_user.id),
        other_settings_form=other_settings_form,
        all_timezones=all_timezones,
        supported_locales=SUPPORTED_LOCALES,
        your_locale=flask.request.accept_languages.best_match(SUPPORTED_LOCALES),
        allowed_language_codes=logic.locale.get_allowed_language_codes(),
    )
    if 'edit_other_settings' in flask.request.form and other_settings_form.validate_on_submit():
        use_schema_editor = flask.request.form.get('input-use-schema-editor', 'yes') != 'no'
        modified_settings: typing.Dict[str, typing.Any] = {
            'USE_SCHEMA_EDITOR': use_schema_editor
        }
        select_timezone = flask.request.form.get('select-timezone', '')
        select_locale = flask.request.form.get('select-locale', '')

        if select_locale == 'auto_lc':
            set_user_settings(
                flask_login.current_user.id,
                {
                    'LOCALE': 'en',
                    'AUTO_LC': True
                }
            )
        elif select_locale in logic.locale.get_allowed_language_codes():
            set_user_settings(flask_login.current_user.id, {'LOCALE': select_locale, 'AUTO_LC': False})

        if not flask.current_app.config['TIMEZONE']:
            if select_timezone == 'auto_tz':
                set_user_settings(flask_login.current_user.id, {'AUTO_TZ': True})
            elif select_timezone in all_timezones:
                set_user_settings(flask_login.current_user.id, {'AUTO_TZ': False, 'TIMEZONE': select_timezone})
            else:
                flask.flash("Invalid timezone", 'error')

        objects_per_page = flask.request.form.get('input-objects-per-page', '')
        if objects_per_page == 'all':
            modified_settings['OBJECTS_PER_PAGE'] = None
        else:
            try:
                modified_settings['OBJECTS_PER_PAGE'] = int(objects_per_page)
            except ValueError:
                pass

        show_object_type_and_id_on_object_page_text = flask.request.form.get('input-show-object-type-and-id-on-object-page', 'default')
        if show_object_type_and_id_on_object_page_text == 'yes':
            show_object_type_and_id_on_object_page = True
        elif show_object_type_and_id_on_object_page_text == 'no':
            show_object_type_and_id_on_object_page = False
        else:
            show_object_type_and_id_on_object_page = None
        modified_settings['SHOW_OBJECT_TYPE_AND_ID_ON_OBJECT_PAGE'] = show_object_type_and_id_on_object_page

        show_object_title_text = flask.request.form.get('input-show-object-title', 'default')
        if show_object_title_text == 'yes':
            show_object_title = True
        elif show_object_title_text == 'no':
            show_object_title = False
        else:
            show_object_title = None
        modified_settings['SHOW_OBJECT_TITLE'] = show_object_title

        full_width_objects_table_text = flask.request.form.get('input-full-width-objects-table', 'default')
        if full_width_objects_table_text == 'yes':
            full_width_objects_table = True
        elif full_width_objects_table_text == 'no':
            full_width_objects_table = False
        else:
            full_width_objects_table = None
        modified_settings['FULL_WIDTH_OBJECTS_TABLE'] = full_width_objects_table

        workflow_view_modals_text = flask.request.form.get('input-workflow-view-modals', 'default')
        if workflow_view_modals_text == 'yes':
            workflow_view_modals = True
        elif workflow_view_modals_text == 'no':
            workflow_view_modals = False
        else:
            workflow_view_modals = None
        modified_settings['WORKFLOW_VIEW_MODALS'] = workflow_view_modals

        workflow_view_collapsed_text = flask.request.form.get('input-workflow-view-collapsed', 'default')
        if workflow_view_collapsed_text == 'yes':
            workflow_view_collapsed = True
        elif workflow_view_collapsed_text == 'no':
            workflow_view_collapsed = False
        else:
            workflow_view_collapsed = None
        modified_settings['WORKFLOW_VIEW_COLLAPSED'] = workflow_view_collapsed

        if flask_login.current_user.is_admin:
            use_admin_permissions = flask.request.form.get('input-use-admin-permissions', 'yes') != 'no'
            modified_settings['USE_ADMIN_PERMISSIONS'] = use_admin_permissions
            show_invitation_log = flask.request.form.get('input-show-invitation-log', 'yes') != 'no'
            modified_settings['SHOW_INVITATION_LOG'] = show_invitation_log
            show_hidden_users_as_admin = flask.request.form.get('input-show-hidden-users-as-admin', 'yes') != 'no'
            modified_settings['SHOW_HIDDEN_USERS_AS_ADMIN'] = show_hidden_users_as_admin
        set_user_settings(flask_login.current_user.id, modified_settings)
        flask_login.current_user.clear_caches()
        refresh()
        flask.flash(lazy_gettext("Successfully updated your settings."), 'success')
        return flask.redirect(flask.url_for('.user_preferences', user_id=flask_login.current_user.id))

    if 'delete_dataverse_api_token' in flask.request.form:
        set_user_settings(flask_login.current_user.id, {
            'DATAVERSE_API_TOKEN': ''
        })
        flask.flash(_('Successfully deleted your stored Dataverse API Token.'), 'success')
        return flask.redirect(flask.url_for('frontend.user_me_preferences'))
    return None


def change_preferences() -> FlaskResponseT:
    template_kwargs: typing.Dict[str, typing.Any] = {}

    response = _handle_account_information_forms(template_kwargs)
    if response is not None:
        return response
    response = _handle_authentication_methods_forms(template_kwargs)
    if response is not None:
        return response
    response = _handle_create_api_token_form(template_kwargs)  # pylint: disable=assignment-from-none
    if response is not None:
        return response
    response = _handle_two_factor_authentication_forms(template_kwargs)
    if response is not None:
        return response
    response = _handle_webhook_forms(template_kwargs)
    if response is not None:
        return response
    response = _handle_notification_forms(template_kwargs)
    if response is not None:
        return response
    response = _handle_default_permissions_forms(template_kwargs)
    if response is not None:
        return response
    response = _handle_other_settings_forms(template_kwargs)
    if response is not None:
        return response

    return flask.render_template(
        'preferences.html',
        **template_kwargs
    )


def confirm_email() -> FlaskResponseT:
    token = flask.request.args.get('token')
    if not token:
        return flask.abort(404)
    data1 = verify_token(token, salt='edit_profile', secret_key=flask.current_app.config['SECRET_KEY'])
    data2 = verify_token(token, salt='add_login', secret_key=flask.current_app.config['SECRET_KEY'])
    if data1 is None and data2 is None:
        return flask.abort(404)
    else:
        if data1 is not None:
            data = data1
            salt = 'edit_profile'
        else:
            data = data2
            salt = 'add_login'
        if isinstance(data, dict) and 'email' in data and 'user_id' in data:
            email = data['email']
            user_id = data['user_id']
        else:
            return flask.abort(400)
        if salt == 'edit_profile':
            logic.users.update_user(
                user_id,
                updating_user_id=user_id,
                email=email
            )
        elif salt == 'add_login':
            try:
                confirm_authentication_method_by_email(
                    user_id=user_id,
                    email=email
                )
            except errors.AuthenticationMethodDoesNotExistError:
                return flask.abort(400)
        else:
            return flask.abort(400)
        db.session.commit()
        user_log.edit_user_preferences(user_id=user_id)
        return flask.redirect(flask.url_for('.user_preferences', user_id=user_id))


def email_for_resetting_password() -> FlaskResponseT:
    request_password_reset_form = RequestPasswordResetForm()
    if flask.request.method == "GET":
        #  GET (email dialog )
        return flask.render_template(
            'reset_password_by_email.html',
            request_password_reset_form=request_password_reset_form
        )
    if flask.request.method == "POST":
        has_error = False
        if request_password_reset_form.validate_on_submit():
            email = request_password_reset_form.email.data
            if '@' not in email:
                has_error = True
            else:
                mail_send_status_info = send_recovery_email(email)
                if mail_send_status_info is not None and mail_send_status_info[0] == BackgroundTaskStatus.FAILED:
                    flask.flash(_("Sending an email failed. Please try again later or contact an administrator."), 'error')
                else:
                    return flask.render_template(
                        'recovery_email_send.html',
                        email=email,
                        has_error=has_error
                    )
        return flask.render_template(
            'reset_password_by_email.html',
            request_password_reset_form=request_password_reset_form,
            has_error=has_error
        )
    return flask.abort(http.HTTPStatus.METHOD_NOT_ALLOWED)


def reset_password() -> FlaskResponseT:
    token = flask.request.args.get('token')
    if not token:
        return flask.abort(404)
    authentication_id = verify_token(token, salt='password', secret_key=flask.current_app.config['SECRET_KEY'])
    if not authentication_id:
        return flask.abort(404)
    authentication_method = get_authentication_method(authentication_method_id=authentication_id)
    password_form = PasswordForm()
    has_error = False
    if authentication_method is None:
        return flask.abort(404)
    username = authentication_method.login['login']
    if flask.request.method == "GET":
        # confirmation dialog
        return flask.render_template('password.html', password_form=password_form, has_error=has_error,
                                     user=authentication_method.user, username=username)
    else:
        if password_form.validate_on_submit():
            if change_password_in_authentication_method(authentication_method.id, password_form.password.data):
                flask.flash(_("Your password has been reset."), 'success')
            else:
                flask.flash(_("Resetting your password failed. Please try again or contact an administrator."), 'error')
            return flask.redirect(flask.url_for('frontend.index'))
        return flask.render_template('password.html', password_form=password_form, has_error=has_error,
                                     user=authentication_method.user, username=username)



import datetime
import enum
import typing

from sqlalchemy.orm import Query, Mapped

from .. import db
from .utils import Model


@enum.unique
class LocationLogEntryType(enum.Enum):
    OTHER = 0
    CREATE_LOCATION = 1
    UPDATE_LOCATION = 2
    ADD_OBJECT = 3
    CHANGE_OBJECT = 4
    REMOVE_OBJECT = 5


class LocationLogEntry(Model):
    __tablename__ = 'location_log_entries'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    type: Mapped[LocationLogEntryType] = db.Column(db.Enum(LocationLogEntryType), nullable=False)
    location_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('locations.id'), nullable=False)
    user_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=True)
    data: Mapped[typing.Dict[str, typing.Any]] = db.Column(db.JSON, nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["LocationLogEntry"]]

    def __init__(
            self,
            location_id: int,
            user_id: typing.Optional[int],
            type: LocationLogEntryType,
            data: typing.Dict[str, typing.Any],
            utc_datetime: typing.Optional[datetime.datetime] = None
    ) -> None:
        super().__init__(
            type=type,
            location_id=location_id,
            user_id=user_id,
            data=data,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc)
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, type={self.type}, location_id={self.location_id}, user_id={self.user_id}, utc_datetime={self.utc_datetime}, data={self.data})>'



# coding: utf-8
"""
Implementation of convert_to_schema(data, previous_schema, new_schema)
"""
import copy
import typing

from flask_babel import _

from .generate_placeholder import generate_placeholder
from .data_diffs import calculate_diff, iter_diff
from .utils import get_dimensionality_for_units
from ..utils import get_translated_text
from ...models import ActionType
from .. import actions, errors, objects


def is_converting_to_schema_necessary(
        data: typing.Dict[str, typing.Any],
        previous_schema: typing.Dict[str, typing.Any],
        new_schema: typing.Dict[str, typing.Any]
) -> typing.Tuple[bool, typing.Sequence[str]]:
    """
    Check whether data is changed by converting it to the new schema.

    :param data: the sampledb object data
    :param previous_schema: the sampledb object schema for the given data
    :param new_schema: the target sampledb object schema
    :return: whether converting the data is necessary and messages about the conversion
    """
    new_data, messages = convert_to_schema(
        data=data,
        previous_schema=previous_schema,
        new_schema=new_schema
    )
    if data == new_data:
        return False, messages
    data_diff = calculate_diff(data_before=data, data_after=new_data)
    messages = list(messages)
    added_property_paths = []
    removed_property_paths = []
    modified_property_paths = []
    for property_path, property_diff in sorted(iter_diff(data_diff)):
        if '_before' in property_diff and '_after' in property_diff:
            modified_property_paths.append(list(property_path))
        elif '_before' in property_diff:
            removed_property_paths.append(list(property_path))
        elif '_after' in property_diff:
            added_property_paths.append(list(property_path))
    if modified_property_paths:
        messages.append(_("Modified properties: %(property_paths)s.", property_paths=', '.join(repr('.'.join(map(str, property_path))) for property_path in modified_property_paths)))
    if added_property_paths:
        messages.append(_("Added properties: %(property_paths)s.", property_paths=', '.join(repr('.'.join(map(str, property_path))) for property_path in added_property_paths)))
    if removed_property_paths:
        messages.append(_("Removed properties: %(property_paths)s.", property_paths=', '.join(repr('.'.join(map(str, property_path))) for property_path in removed_property_paths)))
    return data != new_data, messages


def convert_to_schema(
        data: typing.Dict[str, typing.Any],
        previous_schema: typing.Dict[str, typing.Any],
        new_schema: typing.Dict[str, typing.Any]
) -> typing.Tuple[typing.Any, typing.Sequence[str]]:
    """
    Convert data from one schema to another.

    :param data: the sampledb object data
    :param previous_schema: the sampledb object schema for the given data
    :param new_schema: the target sampledb object schema
    :return: the converted data and a list of conversion warnings/notes
    """
    result: typing.Optional[typing.Tuple[typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[str]]], typing.Sequence[str]]]

    if new_schema == previous_schema and new_schema['type'] in (
            'bool',
            'text',
            'datetime',
            'tags',
            'sample',
            'measurement',
            'object_reference',
            'quantity',
            'array',
            'object',
            'hazards',
            'timeseries',
            'user',
            'file',
            'plotly_chart',
    ):
        return data, []

    if new_schema['type'] == 'tags' and previous_schema['type'] == 'text' and isinstance(data, dict):
        result = _try_convert_text_to_tags(data)
        if result:
            return result
    if new_schema['type'] == 'object_reference' and previous_schema['type'] in ('sample', 'measurement'):
        result = _try_convert_legacy_object_reference_to_object_reference(data, new_schema, previous_schema)
        if result:
            return result
    if previous_schema['type'] == 'object_reference' and new_schema['type'] in ('sample', 'measurement'):
        result = _try_convert_object_reference_to_legacy_object_reference(data, new_schema, previous_schema)
        if result:
            return result
    if previous_schema['type'] != new_schema['type']:
        return copy.deepcopy(generate_placeholder(new_schema)), [_("Unable to convert property '%(title)s' from type '%(type1)s' to type '%(type2)s'.", title=get_translated_text(new_schema['title']), type1=previous_schema['type'], type2=new_schema['type'])]
    if new_schema['type'] == 'text' and 'choices' in new_schema and isinstance(data, dict):
        result = _try_convert_text_to_choices(data, new_schema)
        if result:
            return result
    if new_schema['type'] in ('bool', 'text', 'datetime', 'tags', 'sample', 'measurement', 'hazards', 'user', 'plotly_chart', 'file'):
        return data, []
    if new_schema['type'] == 'object_reference':
        result = _try_convert_object_reference_to_object_reference(data, new_schema, previous_schema)
        if result:
            return result
    if new_schema['type'] == 'quantity':
        result = _try_convert_quantity_to_quantity(data, new_schema, previous_schema)
        if result:
            return result
    if previous_schema['type'] == 'timeseries':
        result = _try_convert_timeseries_to_timeseries(data, new_schema, previous_schema)
        if result:
            return result
    if new_schema['type'] == 'object':
        result = _try_convert_object_to_object(data, new_schema, previous_schema)
        if result:
            return result
    if new_schema['type'] == 'array' and isinstance(data, list):
        result = _try_convert_array_to_array(data, new_schema, previous_schema)
        if result:
            return result
    return copy.deepcopy(generate_placeholder(new_schema)), [_("Unable to convert property '%(title)s' of type '%(type)s'.", title=get_translated_text(new_schema['title']), type=new_schema['type'])]


def _try_convert_text_to_tags(
        data: typing.Dict[str, typing.Any]
) -> typing.Optional[typing.Tuple[typing.Dict[str, typing.Any], typing.Sequence[str]]]:
    tags = []
    for tag in data['text'].split(','):
        tag = tag.strip().lower()
        if tag not in tags:
            tags.append(tag)
    new_data = {
        '_type': 'tags',
        'tags': tags
    }
    return new_data, []


def _try_convert_text_to_choices(
        data: typing.Dict[str, typing.Any],
        new_schema: typing.Dict[str, typing.Any]
) -> typing.Optional[typing.Tuple[typing.Optional[typing.Dict[str, typing.Any]], typing.Sequence[str]]]:
    warning_text = _("Unable to convert property '%(title)s' of type '%(type)s'.", title=get_translated_text(new_schema['title']), type=new_schema['type'])
    if set(data.keys()) != {'_type', 'text'}:
        return None, [warning_text]
    if data['text'] in new_schema['choices']:
        return data, []
    dict_text = data['text'] if isinstance(data['text'], dict) else {'en': data['text']}
    dict_choices = [
        choice if isinstance(choice, dict) else {'en': choice}
        for choice in new_schema['choices']
    ]
    matching_choices = [
        choice_index
        for choice_index, choice in enumerate(dict_choices)
        if all(
            lang_code not in choice or choice[lang_code] == dict_text[lang_code]
            for lang_code in dict_text
        )
    ]
    if len(matching_choices) == 1:
        new_data = copy.deepcopy(data)
        new_data['text'] = new_schema['choices'][matching_choices[0]]
        return new_data, []
    return None, [warning_text]


def _try_convert_object_to_object(
        data: typing.Dict[str, typing.Any],
        new_schema: typing.Dict[str, typing.Any],
        previous_schema: typing.Dict[str, typing.Any]
) -> typing.Optional[typing.Tuple[typing.Dict[str, typing.Any], typing.Sequence[str]]]:
    upgrade_warnings = []
    new_data = copy.deepcopy(generate_placeholder(new_schema))
    if not isinstance(new_data, dict):
        return None
    for property_name, property_value in data.items():
        if property_name in new_schema['properties']:
            new_property_value, property_upgrade_warnings = convert_to_schema(
                data=property_value,
                previous_schema=previous_schema['properties'][property_name],
                new_schema=new_schema['properties'][property_name]
            )
            if new_property_value is not None:
                new_data[property_name] = new_property_value
            for upgrade_warning in property_upgrade_warnings:
                if upgrade_warning not in upgrade_warnings:
                    upgrade_warnings.append(upgrade_warning)
    for property_name in new_schema['properties']:
        # check if any properties were explicitly not set
        if property_name in new_data and property_name not in data and property_name not in new_schema.get('required', []) and property_name in previous_schema['properties']:
            del new_data[property_name]
    return new_data, upgrade_warnings


def _try_convert_array_to_array(
        data: typing.List[typing.Any],
        new_schema: typing.Dict[str, typing.Any],
        previous_schema: typing.Dict[str, typing.Any]
) -> typing.Optional[typing.Tuple[typing.List[typing.Any], typing.Sequence[str]]]:
    new_data = []
    upgrade_warnings = []
    for item in data:
        new_item, item_upgrade_warnings = convert_to_schema(item, previous_schema['items'], new_schema['items'])
        new_data.append(new_item)
        for upgrade_warning in item_upgrade_warnings:
            if upgrade_warning not in upgrade_warnings:
                upgrade_warnings.append(upgrade_warning)
    return new_data, upgrade_warnings


def _try_convert_quantity_to_quantity(
        data: typing.Dict[str, typing.Any],
        new_schema: typing.Dict[str, typing.Any],
        previous_schema: typing.Dict[str, typing.Any]
) -> typing.Optional[typing.Tuple[typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[str]]], typing.Sequence[str]]]:
    previous_dimensionality = get_dimensionality_for_units(previous_schema['units'])
    new_dimensionality = get_dimensionality_for_units(new_schema['units'])
    if new_dimensionality == previous_dimensionality:
        return data, []
    return generate_placeholder(new_schema), [_("Unable to convert quantity '%(title)s' to different dimensionality: %(dimensionality1)s -> %(dimensionality2)s", title=get_translated_text(new_schema['title']), dimensionality1=previous_dimensionality, dimensionality2=new_dimensionality)]


def _try_convert_timeseries_to_timeseries(
        data: typing.Dict[str, typing.Any],
        new_schema: typing.Dict[str, typing.Any],
        previous_schema: typing.Dict[str, typing.Any]
) -> typing.Optional[typing.Tuple[typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.List[str]]], typing.Sequence[str]]]:
    previous_dimensionality = get_dimensionality_for_units(previous_schema['units'])
    new_dimensionality = get_dimensionality_for_units(new_schema['units'])
    if new_dimensionality == previous_dimensionality:
        return data, []
    return generate_placeholder(new_schema), [_("Unable to convert timeseries '%(title)s' to different dimensionality: %(dimensionality1)s -> %(dimensionality2)s", title=get_translated_text(new_schema['title']), dimensionality1=previous_dimensionality, dimensionality2=new_dimensionality)]


def _try_convert_object_reference_to_object_reference(
        data: typing.Dict[str, typing.Any],
        new_schema: typing.Dict[str, typing.Any],
        previous_schema: typing.Dict[str, typing.Any]
) -> typing.Optional[typing.Tuple[typing.Dict[str, typing.Any], typing.Sequence[str]]]:
    referenced_object = None
    action_type_compatible = 'action_type_id' not in new_schema or new_schema['action_type_id'] is None or ('action_type_id' in previous_schema and new_schema['action_type_id'] == previous_schema['action_type_id'])
    if not action_type_compatible and ('action_type_id' not in previous_schema or previous_schema['action_type_id'] is None):
        if 'object_id' in data:
            try:
                referenced_object = objects.get_object(data['object_id'])
                if referenced_object.action_id is None:
                    action_type_id = None
                else:
                    action = actions.get_action(referenced_object.action_id)
                    action_type_id = action.type_id
            except errors.ObjectDoesNotExistError:
                pass
            except errors.ActionDoesNotExistError:
                pass
            else:
                if action_type_id is not None:
                    if isinstance(new_schema['action_type_id'], int) and action_type_id == new_schema['action_type_id']:
                        action_type_compatible = True
                    if isinstance(new_schema['action_type_id'], list) and action_type_id in new_schema['action_type_id']:
                        action_type_compatible = True
        else:
            action_type_compatible = True
    action_compatible = 'action_id' not in new_schema or new_schema['action_id'] is None or ('action_id' in previous_schema and new_schema['action_id'] == previous_schema['action_id'])
    if not action_compatible and ('action_id' not in previous_schema or previous_schema['action_id'] is None):
        if 'object_id' in data:
            try:
                if referenced_object is None:
                    referenced_object = objects.get_object(data['object_id'])
                action_id = referenced_object.action_id
            except errors.ObjectDoesNotExistError:
                pass
            else:
                if action_id is not None:
                    if isinstance(new_schema['action_id'], int) and action_id == new_schema['action_id']:
                        action_compatible = True
                    if isinstance(new_schema['action_id'], list) and action_id in new_schema['action_id']:
                        action_compatible = True
        else:
            action_compatible = True
    if action_type_compatible and action_compatible:
        return data, []
    if (action_type_compatible or action_compatible) and new_schema.get('filter_operator', 'and') == 'or':
        return data, []
    return None


def _try_convert_object_reference_to_legacy_object_reference(
        data: typing.Dict[str, typing.Any],
        new_schema: typing.Dict[str, typing.Any],
        previous_schema: typing.Dict[str, typing.Any]
) -> typing.Optional[typing.Tuple[typing.Dict[str, typing.Any], typing.Sequence[str]]]:
    if 'action_type_id' in previous_schema:
        if new_schema['type'] == 'sample' and previous_schema['action_type_id'] == ActionType.SAMPLE_CREATION:
            return data, []
        if new_schema['type'] == 'measurement' and previous_schema['action_type_id'] == ActionType.MEASUREMENT:
            return data, []
    if 'action_type_id' not in previous_schema or previous_schema['action_type_id'] is None:
        if 'object_id' in data:
            try:
                referenced_object = objects.get_object(data['object_id'])
                if referenced_object.action_id is None:
                    action_type_id = None
                else:
                    action = actions.get_action(referenced_object.action_id)
                    action_type_id = action.type_id
            except errors.ObjectDoesNotExistError:
                pass
            except errors.ActionDoesNotExistError:
                pass
            else:
                if action_type_id is not None:
                    if new_schema['type'] == 'sample' and action_type_id == ActionType.SAMPLE_CREATION:
                        return data, []
                    if new_schema['type'] == 'measurement' and action_type_id == ActionType.MEASUREMENT:
                        return data, []
    return None


def _try_convert_legacy_object_reference_to_object_reference(
        data: typing.Dict[str, typing.Any],
        new_schema: typing.Dict[str, typing.Any],
        previous_schema: typing.Dict[str, typing.Any]
) -> typing.Optional[typing.Tuple[typing.Dict[str, typing.Any], typing.Sequence[str]]]:
    action_type_compatible = False
    if 'action_type_id' in new_schema and new_schema['action_type_id'] is not None:
        if previous_schema['type'] == 'sample' and new_schema['action_type_id'] == ActionType.SAMPLE_CREATION:
            action_type_compatible = True
        if previous_schema['type'] == 'measurement' and new_schema['action_type_id'] == ActionType.MEASUREMENT:
            action_type_compatible = True
    else:
        action_type_compatible = True
    action_compatible = False
    if 'action_id' in new_schema and new_schema['action_id'] is not None:
        if 'object_id' in data:
            try:
                referenced_object = objects.get_object(data['object_id'])
                action_id = referenced_object.action_id
            except errors.ObjectDoesNotExistError:
                pass
            else:
                if action_id == new_schema['action_id']:
                    action_compatible = True
        else:
            action_compatible = True
    else:
        action_compatible = True
    if action_type_compatible and action_compatible:
        return {
            '_type': 'object_reference',
            'object_id': data['object_id']
        }, []
    if (action_type_compatible or action_compatible) and new_schema.get('filter_operator', 'and') == 'or':
        return {
            '_type': 'object_reference',
            'object_id': data['object_id']
        }, []
    return None



import copy
import re
import typing

from .. import errors
from ..datatypes import JSONEncoder
from .validate import validate

GenericDiff = typing.TypedDict('GenericDiff', {'_before': typing.Any, '_after': typing.Any}, total=False)
ArrayDiff = typing.List[typing.Optional['DataDiff']]
ArrayIndexDiff = typing.Dict[str, GenericDiff]
ObjectDiff = typing.Dict[str, typing.Optional['DataDiff']]
DataDiff = typing.Union[ArrayDiff, ArrayIndexDiff, ObjectDiff, GenericDiff]


class _ValueNotSet:
    pass


VALUE_NOT_SET = _ValueNotSet()


def _calculate_object_diff(data_before: typing.Dict[str, typing.Any], data_after: typing.Dict[str, typing.Any]) -> ObjectDiff:
    data_diff: ObjectDiff = {}
    for property_name in set(data_before) | set(data_after):
        value_before = data_before.get(property_name, VALUE_NOT_SET)
        value_after = data_after.get(property_name, VALUE_NOT_SET)
        value_diff = calculate_diff(value_before, value_after)
        if value_diff is not None:
            data_diff[property_name] = value_diff
    return data_diff


def _calculate_array_diff(data_before: typing.List[typing.Any], data_after: typing.List[typing.Any]) -> ArrayDiff:
    length_before = len(data_before)
    length_after = len(data_after)
    length_diff = max(length_before, length_after)
    data_diff: ArrayDiff = [None] * length_diff
    for index in range(length_diff):
        value_before = data_before[index] if index < length_before else VALUE_NOT_SET
        value_after = data_after[index] if index < length_after else VALUE_NOT_SET
        value_diff = calculate_diff(value_before, value_after)
        if value_diff is not None:
            data_diff[index] = value_diff
    return data_diff


def calculate_diff(data_before: typing.Any, data_after: typing.Any) -> typing.Optional[DataDiff]:
    if data_before == data_after:
        return None
    type_before = _guess_type_of_data(data_before)
    type_after = _guess_type_of_data(data_after)
    if type_before == type_after:
        if type_before == 'array':
            return _calculate_array_diff(data_before, data_after)
        if type_before == 'object':
            return _calculate_object_diff(data_before, data_after)
    data_diff: GenericDiff = {}
    if data_before != VALUE_NOT_SET:
        data_diff['_before'] = data_before
    if data_after != VALUE_NOT_SET:
        data_diff['_after'] = data_after
    return data_diff


def _guess_type_of_data(data: typing.Any) -> typing.Optional[str]:
    if isinstance(data, list):
        return 'array'
    if isinstance(data, dict):
        return str(data.get('_type', 'object'))
    return None


def _apply_array_diff(
        data_before: typing.List[typing.Any],
        data_diff: ArrayDiff,
        schema_before: typing.Optional[typing.Dict[str, typing.Any]]
) -> typing.List[typing.Any]:
    data_after = copy.deepcopy(data_before)
    indices_to_remove = []
    for index, item_diff in enumerate(data_diff):
        value_before = data_before[index] if index < len(data_before) else VALUE_NOT_SET
        if schema_before is None:
            item_schema_before = None
        else:
            try:
                item_schema_before = schema_before['items']
            except Exception:
                raise errors.DiffMismatchError()
        value_after = apply_diff(value_before, item_diff, item_schema_before, validate_data_before=False)
        if value_after != VALUE_NOT_SET:
            if value_before == VALUE_NOT_SET:
                data_after.append(value_after)
            else:
                data_after[index] = value_after
        elif value_before != VALUE_NOT_SET:
            indices_to_remove.append(index)
    for index in reversed(indices_to_remove):
        del data_after[index]
    return data_after


def _apply_timeseries_array_diff(
        data_before: typing.Dict[str, typing.Any],
        data_diff: ArrayDiff
) -> typing.Dict[str, typing.Any]:
    if not isinstance(data_before.get('data'), list):
        raise errors.DiffMismatchError()
    data_after = copy.deepcopy(data_before)
    data_after['data'] = _apply_array_diff(data_before=data_before['data'], data_diff=data_diff, schema_before=None)
    return data_after


def _apply_array_index_diff(
        data_before: typing.List[typing.Any],
        data_diff: ArrayIndexDiff,
        schema_before: typing.Optional[typing.Dict[str, typing.Any]]
) -> typing.List[typing.Any]:
    length_before = len(data_before)
    diffs_by_index = {}
    for index_info, item_diff in data_diff.items():
        try:
            index = int(index_info)
        except ValueError:
            raise errors.DiffMismatchError()
        if index_info.startswith('+') or index < 0:
            index += length_before
        if index < 0:
            raise errors.DiffMismatchError()
        if index in diffs_by_index:
            raise errors.DiffMismatchError()
        diffs_by_index[index] = item_diff
    new_indices = {index for index in diffs_by_index if index >= length_before}
    if new_indices:
        length_after = max(new_indices) + 1
        if new_indices != set(range(length_before, length_after)):
            raise errors.DiffMismatchError()
    else:
        length_after = length_before
    new_data_diff: ArrayDiff = [
        diffs_by_index.get(index)
        for index in range(length_after)
    ]
    return _apply_array_diff(data_before, new_data_diff, schema_before)


def _apply_timeseries_array_index_diff(
        data_before: typing.Dict[str, typing.Any],
        data_diff: ArrayIndexDiff
) -> typing.Dict[str, typing.Any]:
    if not isinstance(data_before.get('data'), list):
        raise errors.DiffMismatchError()
    data_after = copy.deepcopy(data_before)
    data_after['data'] = _apply_array_index_diff(data_before=data_before['data'], data_diff=data_diff, schema_before=None)
    return data_after


def _apply_object_diff(
        data_before: typing.Dict[str, typing.Any],
        data_diff: ObjectDiff,
        schema_before: typing.Optional[typing.Dict[str, typing.Any]]
) -> typing.Dict[str, typing.Any]:
    if schema_before is None:
        schema_before = {
            'title': '',
            'type': 'object',
            'properties': {}
        }
    data_after = copy.deepcopy(data_before)
    for property_name, property_diff in data_diff.items():
        value_before = data_before.get(property_name, VALUE_NOT_SET)
        try:
            property_schema_before = schema_before['properties'][property_name]
        except Exception:
            if isinstance(property_diff, dict) and '_after' in property_diff and '_before' not in property_diff:
                # the property did not exist before in this specific case, so an empty schema can be used
                property_schema_before = {}
            else:
                raise errors.DiffMismatchError()
        value_after = apply_diff(value_before, property_diff, property_schema_before, validate_data_before=False)
        if value_after != VALUE_NOT_SET:
            data_after[property_name] = value_after
        elif value_before != VALUE_NOT_SET:
            del data_after[property_name]
    return data_after


def _compare_generic_data(
        data_left: typing.Any,
        data_right: typing.Any
) -> bool:
    if data_left is None and data_right is None:
        return True
    if data_left == data_right:
        return True
    if isinstance(data_left, dict) and isinstance(data_right, dict):
        if '_type' in data_left and '_type' in data_right:
            if data_left['_type'] != data_right['_type']:
                return False
            try:
                value_left = JSONEncoder.serializable_types[data_left['_type']].from_json(data_left)
                value_right = JSONEncoder.serializable_types[data_right['_type']].from_json(data_right)
                return bool(value_left == value_right)
            except Exception:
                return False
        else:
            keys_left = set(data_left)
            keys_right = set(data_right)
            if keys_left != keys_right:
                return False
            return all(
                _compare_generic_data(data_left[key], data_right[key])
                for key in keys_left
            )
    if isinstance(data_left, list) and isinstance(data_right, list):
        if len(data_left) != len(data_right):
            return False
        return all(
            _compare_generic_data(item_left, item_right)
            for item_left, item_right in zip(data_left, data_right)
        )
    return False


def _apply_generic_diff(
        data_before: typing.Any,
        data_diff: GenericDiff,
        schema_before: typing.Optional[typing.Dict[str, typing.Any]],
        *,
        validate_data_before: bool = True
) -> typing.Any:
    if '_before' in data_diff:
        if data_diff['_before'] is not None and schema_before is not None and validate_data_before:
            try:
                validate(data_diff['_before'], schema_before)
            except Exception:
                raise errors.DiffMismatchError()
        if not _compare_generic_data(data_before, data_diff['_before']):
            raise errors.DiffMismatchError()
    else:
        if data_before != VALUE_NOT_SET:
            raise errors.DiffMismatchError()
    if '_after' in data_diff:
        return copy.deepcopy(data_diff['_after'])
    else:
        if data_before == VALUE_NOT_SET:
            raise errors.DiffMismatchError()
        return VALUE_NOT_SET


def _guess_type_of_diff(data_diff: typing.Optional[DataDiff]) -> typing.Optional[typing.Union[typing.Type[ArrayDiff], typing.Type[ArrayIndexDiff], typing.Type[ObjectDiff], typing.Type[GenericDiff]]]:
    if data_diff is None:
        return None
    if isinstance(data_diff, list):
        return ArrayDiff
    assert isinstance(data_diff, dict)
    if all(type(key) is str and re.match(r'^([+-]?[1-9][0-9]*)|(\+?0)$', key) for key in data_diff):
        return ArrayIndexDiff
    if '_before' not in data_diff and '_after' not in data_diff:
        return ObjectDiff
    return GenericDiff


def apply_diff(
        data_before: typing.Any,
        data_diff: typing.Optional[DataDiff],
        schema_before: typing.Optional[typing.Dict[str, typing.Any]],
        *,
        validate_data_before: bool = True
) -> typing.Any:
    if validate_data_before and schema_before is not None:
        validate(data_before, schema_before)
    diff_type = _guess_type_of_diff(data_diff)
    if diff_type is None:
        if data_before == VALUE_NOT_SET:
            raise errors.DiffMismatchError()
        return copy.deepcopy(data_before)
    if diff_type is ArrayDiff:
        if isinstance(data_before, dict) and data_before.get('_type') == 'timeseries':
            return _apply_timeseries_array_diff(data_before, typing.cast(ArrayDiff, data_diff))
        if isinstance(data_before, list):
            return _apply_array_diff(data_before, typing.cast(ArrayDiff, data_diff), schema_before)
        raise errors.DiffMismatchError()
    if diff_type is ArrayIndexDiff:
        if isinstance(data_before, dict) and data_before.get('_type') == 'timeseries':
            return _apply_timeseries_array_index_diff(data_before, typing.cast(ArrayIndexDiff, data_diff))
        if isinstance(data_before, list):
            return _apply_array_index_diff(data_before, typing.cast(ArrayIndexDiff, data_diff), schema_before)
        raise errors.DiffMismatchError()
    if diff_type is ObjectDiff:
        if not isinstance(data_before, dict):
            raise errors.DiffMismatchError()
        return _apply_object_diff(data_before, typing.cast(ObjectDiff, data_diff), schema_before)
    if diff_type is GenericDiff:
        return _apply_generic_diff(data_before, typing.cast(GenericDiff, data_diff), schema_before, validate_data_before=validate_data_before)
    return VALUE_NOT_SET


def invert_diff(
        data_diff: typing.Optional[DataDiff]
) -> typing.Optional[DataDiff]:
    diff_type = _guess_type_of_diff(data_diff)
    if diff_type is ArrayDiff:
        array_diff = typing.cast(ArrayDiff, data_diff)
        return [
            invert_diff(data_diff=item_diff)
            for item_diff in array_diff
        ]
    if diff_type is ArrayIndexDiff:
        array_index_diff = typing.cast(ArrayIndexDiff, data_diff)
        inverted_data_diff: ArrayIndexDiff = {}
        length_change = 0
        for item_diff in array_index_diff.values():
            if not isinstance(item_diff, dict):
                continue
            if '_after' in item_diff and '_before' not in item_diff:
                length_change += 1
            if '_before' in item_diff and '_after' not in item_diff:
                length_change -= 1
        for index_info, item_diff in array_index_diff.items():
            inverted_item_diff = typing.cast(GenericDiff, invert_diff(data_diff=item_diff))
            if index_info.startswith('+') or index_info.startswith('-'):
                try:
                    index_offset = int(index_info)
                except ValueError:
                    raise errors.DiffMismatchError()
                inverted_data_diff[f"{index_offset - length_change:+d}"] = inverted_item_diff
            else:
                inverted_data_diff[index_info] = inverted_item_diff
        return inverted_data_diff
    if diff_type is ObjectDiff:
        object_diff = typing.cast(ObjectDiff, data_diff)
        return {
            property_name: invert_diff(data_diff=property_diff)
            for property_name, property_diff in object_diff.items()
        }
    if diff_type is GenericDiff:
        generic_diff = typing.cast(GenericDiff, data_diff)
        inverted_data_diff = {}
        if '_before' in generic_diff:
            inverted_data_diff['_after'] = generic_diff['_before']
        if '_after' in generic_diff:
            inverted_data_diff['_before'] = generic_diff['_after']
        return inverted_data_diff
    return None


def iter_diff(
        data_diff: typing.Optional[DataDiff],
        property_path: typing.Sequence[typing.Union[int, str]] = ()
) -> typing.Generator[typing.Tuple[typing.Sequence[typing.Union[int, str]], GenericDiff], None, None]:
    diff_type = _guess_type_of_diff(data_diff)
    if diff_type is None:
        return
    if diff_type is ArrayDiff:
        array_diff = typing.cast(ArrayDiff, data_diff)
        for index, item_diff in enumerate(array_diff):
            yield from iter_diff(item_diff, tuple(property_path) + (index,))
    if diff_type is ArrayIndexDiff:
        array_index_diff = typing.cast(ArrayIndexDiff, data_diff)
        for index_info, item_diff in array_index_diff.items():
            yield from iter_diff(item_diff, tuple(property_path) + (index_info,))
    if diff_type is ObjectDiff:
        object_diff = typing.cast(ObjectDiff, data_diff)
        for property_name, property_diff in object_diff.items():
            yield from iter_diff(property_diff, tuple(property_path) + (property_name,))
        return
    if diff_type is GenericDiff:
        yield property_path, typing.cast(GenericDiff, data_diff)
    return



# coding: utf-8
"""
Logic module for management of action type translations

Action type translations contain all language-dependent attributes of an
action type for a specific language.
"""

import dataclasses
import typing

from .. import db
from . import errors, languages, action_types
from ..logic.languages import Language
from .. import models


@dataclasses.dataclass(frozen=True)
class ActionTypeTranslation:
    """
    This class provides an immutable wrapper around models.action_translations.ActionTypeTranslation.
    """
    action_type_id: int
    language_id: int
    name: str
    description: str
    object_name: str
    object_name_plural: str
    view_text: str
    perform_text: str
    _language_cache: typing.List[languages.Language] = dataclasses.field(default_factory=list, kw_only=True, repr=False, compare=False)

    @classmethod
    def from_database(cls, action_type_translation: models.ActionTypeTranslation) -> 'ActionTypeTranslation':
        return ActionTypeTranslation(
            action_type_id=action_type_translation.action_type_id,
            language_id=action_type_translation.language_id,
            name=action_type_translation.name,
            description=action_type_translation.description,
            object_name=action_type_translation.object_name,
            object_name_plural=action_type_translation.object_name_plural,
            view_text=action_type_translation.view_text,
            perform_text=action_type_translation.perform_text
        )

    @property
    def language(self) -> Language:
        if not self._language_cache:
            self._language_cache.append(languages.get_language(self.language_id))
        return self._language_cache[0]


def get_action_type_translations_for_action_type(
        action_type_id: int,
        use_fallback: bool = False
) -> typing.List[ActionTypeTranslation]:
    """
    Return all translations for the action type with the given action type ID.

    :param action_type_id: the ID of an existing action type
    :param use_fallback: whether a fallback translation may be returned
    :return: a list containing all action type translations for the given action type
    :raise errors.ActionTypeDoesNotExistError: when no action type with the
        given action type ID exists
    :raise errors.ActionTypeTranslationDoesNotExistError: when no action type translation exists
        for the given action type
    """
    action_type_translations = models.ActionTypeTranslation.query.filter_by(
        action_type_id=action_type_id
    ).order_by(models.ActionTypeTranslation.language_id).all()
    if not action_type_translations and use_fallback:
        return [
            ActionTypeTranslation(
                action_type_id=action_type_id,
                language_id=Language.ENGLISH,
                name=f'#{action_type_id}',
                description='',
                object_name='Object',
                object_name_plural='Objects',
                view_text='View Objects',
                perform_text='Create Object'
            )
        ]
    return [
        ActionTypeTranslation.from_database(action_type_translation)
        for action_type_translation in action_type_translations
    ]


def set_action_type_translation(
        action_type_id: int,
        language_id: int,
        name: str,
        description: str,
        object_name: str,
        object_name_plural: str,
        view_text: str,
        perform_text: str
) -> ActionTypeTranslation:
    """
    Create or update an action type translation.

    :param action_type_id: the ID of the action type
    :param language_id: the ID of the language
    :param name: the name of the action type
    :param description: the description of the action type
    :param object_name: the name of an object created by an action of this type
    :param object_name_plural: the plural form of the object name
    :param view_text: the text to display for viewing objects created by actions of this type
    :param perform_text: the text to display for performing actions of this type
    :return: the action type translation
    :raise errors.LanguageDoesNotExistError: if no language with the given ID
        exists
    :raise errors.ActionTypeDoesNotExistError: if no action type with the
        given ID exists
    """
    action_type_translation = models.ActionTypeTranslation.query.filter_by(
        action_type_id=action_type_id,
        language_id=language_id
    ).first()
    if action_type_translation is None:
        action_types.check_action_type_exists(action_type_id)
        languages.get_language(language_id)
        action_type_translation = models.ActionTypeTranslation(
            action_type_id=action_type_id,
            language_id=language_id,
            name=name,
            description=description,
            object_name=object_name,
            object_name_plural=object_name_plural,
            view_text=view_text,
            perform_text=perform_text
        )
    else:
        action_type_translation.name = name
        action_type_translation.description = description
        action_type_translation.object_name = object_name
        action_type_translation.object_name_plural = object_name_plural
        action_type_translation.view_text = view_text
        action_type_translation.perform_text = perform_text
    db.session.add(action_type_translation)
    db.session.commit()
    return ActionTypeTranslation.from_database(action_type_translation)


def delete_action_type_translation(
        action_type_id: int,
        language_id: int
) -> None:
    """
    Deletes the action type translation with the given action type ID and language ID

    :param action_type_id: the ID of an existing action type
    :param language_id: the ID of an existing language
    :raise errors.ActionTypeTranslationDoesNotExistError: when no action type translation exist for the given
        action type and language ID
    """
    action_type_translation = models.ActionTypeTranslation.query.filter_by(
        action_type_id=action_type_id,
        language_id=language_id
    ).first()
    if action_type_translation is None:
        raise errors.ActionTypeTranslationDoesNotExistError()
    db.session.delete(action_type_translation)
    db.session.commit()



# coding: utf-8
"""
Replace NOT NULL constraints per column by NOT NULL constraints conditioned by federation reference in comments.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    constraints = db.session.execute(db.text("""
         SELECT conname
         FROM pg_catalog.pg_constraint
         WHERE conname = 'comments_not_null_check'
    """)).fetchall()

    if len(constraints) > 0:
        return False

    # Perform migration
    db.session.execute(db.text("""
            ALTER TABLE comments
                ADD CONSTRAINT comments_not_null_check
                    CHECK (
                        (
                            fed_id IS NOT NULL AND
                            component_id IS NOT NULL
                        ) OR (
                            user_id IS NOT NULL AND
                            utc_datetime IS NOT NULL
                        )
                    ),
                ALTER COLUMN user_id DROP NOT NULL,
                ALTER COLUMN utc_datetime DROP NOT NULL
        """))
    return True



# coding: utf-8
"""

"""

from .. import db
from .users import User
from .actions import Action
from .components import Component
from .eln_imports import ELNImport
from .versioned_json_object_tables import VersionedJSONSerializableObjectTables, Object

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'

__all__ = [
    'Object',
    'Objects'
]


Objects = VersionedJSONSerializableObjectTables(
    'objects',
    user_id_column=User.id,
    action_id_column=Action.id,
    action_schema_column=Action.schema,
    component_id_column=Component.id,
    eln_import_id_column=ELNImport.id,
    metadata=db.metadata
)



# coding: utf-8
"""
API for data exchange in SampleDB federations
"""

from flask import Blueprint
from .federation import UpdateHook, ImportStatus, Objects, Users, File, Components, FederatedLoginMetadata

federation_api = Blueprint('federation_api', __name__)
federation_api.add_url_rule('/federation/v1/hooks/update/', endpoint='hooks_update', view_func=UpdateHook.as_view('update_hook'))
federation_api.add_url_rule('/federation/v1/shares/objects/<int:object_id>/import_status', endpoint='import_status', view_func=ImportStatus.as_view('import_status'))
federation_api.add_url_rule('/federation/v1/shares/objects/', endpoint='object_updates', view_func=Objects.as_view('objects'))
federation_api.add_url_rule('/federation/v1/shares/users/', endpoint='users', view_func=Users.as_view('users'))
federation_api.add_url_rule('/federation/v1/shares/objects/<int:object_id>/files/<int:file_id>', endpoint='file', view_func=File.as_view('file'))
federation_api.add_url_rule('/federation/v1/shares/components/', endpoint='components', view_func=Components.as_view('components'))
federation_api.add_url_rule('/federation/v1/shares/metadata/', endpoint='federated_login_metadata', view_func=FederatedLoginMetadata.as_view('federated_login_metadata'))



"""
Add the status column to the dataverse_exports table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('dataverse_exports', 'status'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE dataverse_exports
        ALTER COLUMN dataverse_url DROP NOT NULL
    """))
    db.session.execute(db.text("""
        ALTER TABLE dataverse_exports
        ADD COLUMN status dataverseexportstatus NULL
    """))
    db.session.execute(db.text("""
        UPDATE dataverse_exports
        SET status = 'EXPORT_FINISHED'::dataverseexportstatus
    """))
    db.session.execute(db.text("""
        ALTER TABLE dataverse_exports
        ALTER COLUMN status SET NOT NULL
    """))
    return True




import datetime
import typing

from sqlalchemy.orm import Query, Mapped, relationship

from .. import db
from .utils import Model

if typing.TYPE_CHECKING:
    from .users import User


class TemporaryFile(Model):
    __tablename__ = 'temporary_files'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    context_id: Mapped[str] = db.Column(db.String, nullable=False)
    file_name: Mapped[str] = db.Column(db.String, nullable=False)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)
    binary_data: Mapped[bytes] = db.deferred(db.Column(db.LargeBinary, nullable=False))
    uploader: Mapped['User'] = relationship('User')

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["TemporaryFile"]]



# coding: utf-8
"""
Replace the type column by the type_id column in the actions table.
"""

import flask_sqlalchemy

from .utils import table_has_column
from ..actions import ActionType


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('actions', 'type_id'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE actions
        ADD type_id INTEGER NULL
    """))

    default_action_types = [
        {
            'type_id': ActionType.SAMPLE_CREATION,
            'type': 'SAMPLE_CREATION'
        },
        {
            'type_id': ActionType.MEASUREMENT,
            'type': 'MEASUREMENT'
        },
        {
            'type_id': ActionType.SIMULATION,
            'type': 'SIMULATION',
        }
    ]
    for action_type in default_action_types:
        db.session.execute(db.text("""
            UPDATE actions
            SET type_id = :type_id
            WHERE type::text = :type
        """), params=action_type)

    db.session.execute(db.text("""
        ALTER TABLE actions
        ALTER type_id SET NOT NULL
    """))

    db.session.execute(db.text("""
        ALTER TABLE actions
        ADD CONSTRAINT fk_actions_type_id FOREIGN KEY (type_id) REFERENCES action_types (id)
    """))

    db.session.execute(db.text("""
        ALTER TABLE actions
        DROP type
    """))
    return True



# coding: utf-8
"""

"""

import enum
import typing

from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import relationship, Query, Mapped

from .. import db
from .utils import Model

if typing.TYPE_CHECKING:
    from .components import Component


@enum.unique
class ComponentAuthenticationType(enum.Enum):
    TOKEN = 1


class ComponentAuthentication(Model):
    __tablename__ = "component_authentications"

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    component_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('components.id'), nullable=False)
    login: Mapped[typing.Dict[str, typing.Any]] = db.Column(postgresql.JSONB, nullable=False)
    type: Mapped[ComponentAuthenticationType] = db.Column(db.Enum(ComponentAuthenticationType), nullable=False)
    component: Mapped['Component'] = relationship('Component')

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["ComponentAuthentication"]]

    def __init__(
            self,
            login: typing.Dict[str, typing.Any],
            authentication_type: ComponentAuthenticationType,
            component_id: int
    ) -> None:
        super().__init__(
            component_id=component_id,
            login=login,
            type=authentication_type
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id})>'


class OwnComponentAuthentication(Model):
    __tablename__ = "own_component_authentications"

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    component_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('components.id'), nullable=False)
    login: Mapped[typing.Dict[str, typing.Any]] = db.Column(postgresql.JSONB, nullable=False)
    type: Mapped[ComponentAuthenticationType] = db.Column(db.Enum(ComponentAuthenticationType), nullable=False)
    component: Mapped['Component'] = relationship('Component')

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["OwnComponentAuthentication"]]

    def __init__(
            self,
            login: typing.Dict[str, typing.Any],
            authentication_type: ComponentAuthenticationType,
            component_id: int
    ) -> None:
        super().__init__(
            component_id=component_id,
            login=login,
            type=authentication_type
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id})>'



# coding: utf-8
"""

"""
import typing

from sqlalchemy.orm import Mapped, Query

from .. import db
from . import Objects
from .utils import Model


class ObjectPublication(Model):
    __tablename__ = 'object_publications'

    object_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Objects.object_id_column), nullable=False)
    doi: Mapped[str] = db.Column(db.String, nullable=False)
    title: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True)
    object_name: Mapped[typing.Optional[str]] = db.Column(db.String, nullable=True)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["ObjectPublication"]]

    __table_args__ = (
        db.PrimaryKeyConstraint(object_id, doi),
    )



"""

"""
import typing

import flask
from flask import request

from . import frontend
from .objects.permissions import on_unauthorized
from .. import db
from ..logic import errors
from ..logic.files import get_file
from ..logic.security_tokens import generate_token
from ..models import DownloadServiceJobFile, Permissions
from ..utils import object_permissions_required, FlaskResponseT


def upload_file_list(object_id: int) -> typing.Optional[int]:
    job_id = None
    file_ids = request.args.getlist('file_ids', type=int)
    for file_id in file_ids:
        try:
            file = get_file(file_id=int(file_id), object_id=object_id)
        except errors.FileDoesNotExistError:
            continue
        if file.storage != 'local_reference' or file.data is None or not file.data.get('valid'):
            continue
        ds = DownloadServiceJobFile(job_id, object_id, int(file_id))
        db.session.add(ds)
        db.session.commit()
        if job_id is None:
            job_id = ds.id
    return job_id


@frontend.route('/objects/<int:object_id>/download_service/', methods=['GET'])
@object_permissions_required(Permissions.READ, on_unauthorized=on_unauthorized)
def download_service(object_id: int) -> FlaskResponseT:
    download_service_enabled = flask.current_app.config['DOWNLOAD_SERVICE_URL'] and flask.current_app.config['DOWNLOAD_SERVICE_SECRET']
    if not download_service_enabled:
        return flask.abort(404)

    job_id = upload_file_list(object_id)
    if job_id is None:
        return flask.abort(404)
    signature = generate_token(
        job_id,
        'download_service_zipping',
        flask.current_app.config['DOWNLOAD_SERVICE_SECRET']
    )
    url = f'{flask.current_app.config["DOWNLOAD_SERVICE_URL"]}?signature={signature}'
    return flask.redirect(url, code=302)



# coding: utf-8
"""
Helper functions for conditional properties

New condition types also need to be implemented in JavaScript in:
sampledb/static/sampledb/js/conditional_wrapper.js
"""

import typing

from ..errors import ValidationError


def validate_condition_schema(
        condition: typing.Dict[str, typing.Any],
        property_schemas: typing.Dict[str, typing.Any],
        path: typing.List[str]
) -> None:
    """
    Validate a condition from a property schema.

    :param condition: the condition to validate
    :param property_schemas: the schemas of all properties of the parent object
    :param path: the path to the condition
    :raise ValidationError: if the condition is invalid
    """
    if condition['type'] == 'choice_equals':
        if set(condition.keys()) != {'type', 'property_name', 'choice'}:
            raise ValidationError('expected type, property_name and choice in condition', path)
        if condition['property_name'] not in property_schemas:
            raise ValidationError('unknown property_name', path)
        if property_schemas[condition['property_name']].get('type') != 'text' or (condition['choice'] is not None and condition['choice'] not in property_schemas[condition['property_name']].get('choices', [])):
            raise ValidationError('unknown choice', path)
    elif condition['type'] == 'user_equals':
        if set(condition.keys()) != {'type', 'property_name', 'user_id'}:
            raise ValidationError('expected type, property_name and user_id in condition', path)
        if condition['property_name'] not in property_schemas:
            raise ValidationError('unknown property_name', path)
        if property_schemas[condition['property_name']].get('type') != 'user':
            raise ValidationError('property_name does not belong to a user property', path)
        if condition['user_id'] is not None and not isinstance(condition['user_id'], int):
            raise ValidationError('user_id must be integer or None', path)
    elif condition['type'] == 'bool_equals':
        if set(condition.keys()) != {'type', 'property_name', 'value'}:
            raise ValidationError('expected type, property_name and value in condition', path)
        if condition['property_name'] not in property_schemas:
            raise ValidationError('unknown property_name', path)
        if property_schemas[condition['property_name']].get('type') != 'bool':
            raise ValidationError('property_name does not belong to a bool property', path)
        if not isinstance(condition['value'], bool):
            raise ValidationError('value must be boolean', path)
    elif condition['type'] == 'object_equals':
        if set(condition.keys()) != {'type', 'property_name', 'object_id'}:
            raise ValidationError('expected type, property_name and object_id in condition', path)
        if condition['property_name'] not in property_schemas:
            raise ValidationError('unknown property_name', path)
        if property_schemas[condition['property_name']].get('type') not in {'object_reference', 'sample', 'measurement'}:
            raise ValidationError('property_name does not belong to a object_reference, sample or measurement property', path)
        if condition['object_id'] is not None and not isinstance(condition['object_id'], int):
            raise ValidationError('object_id must be integer or None', path)
    elif condition['type'] in ('any', 'all'):
        if set(condition.keys()) != {'type', 'conditions'}:
            raise ValidationError('expected type and conditions in condition', path)
        if not isinstance(condition['conditions'], list):
            raise ValidationError('conditions must be list', path)
        for i, sub_condition in enumerate(condition['conditions']):
            validate_condition_schema(sub_condition, property_schemas, path + [str(i)])
    elif condition['type'] == 'not':
        if set(condition.keys()) != {'type', 'condition'}:
            raise ValidationError('expected type and condition in condition', path)
        validate_condition_schema(condition['condition'], property_schemas, path + ['condition'])
    else:
        raise ValidationError('unknown condition type', path)


def is_condition_fulfilled(
        condition: typing.Dict[str, typing.Any],
        instance: typing.Dict[str, typing.Any]
) -> bool:
    """
    Check whether a particular condition is fulfilled by the parent object.

    :param condition: the condition to check
    :param instance: the instance to check the condition on
    :return: whether or not the condition is fulfilled
    """
    if condition['type'] == 'choice_equals':
        return (
            condition['choice'] is None and
            condition['property_name'] not in instance
        ) or (
            condition['choice'] is not None and
            condition['property_name'] in instance and
            isinstance(instance[condition['property_name']], dict) and
            (
                instance[condition['property_name']].get('text') == condition['choice'] or
                instance[condition['property_name']].get('text') == {'en': condition['choice']} or
                {'en': instance[condition['property_name']].get('text')} == condition['choice']
            )
        )
    if condition['type'] == 'user_equals':
        return (
            condition['user_id'] is None and
            condition['property_name'] not in instance
        ) or (
            condition['user_id'] is not None and
            condition['property_name'] in instance and
            isinstance(instance[condition['property_name']], dict) and
            instance[condition['property_name']].get('user_id') == condition['user_id']
        )
    if condition['type'] == 'bool_equals':
        return (
            condition['value'] is not None and
            condition['property_name'] in instance and
            isinstance(instance[condition['property_name']], dict) and
            instance[condition['property_name']].get('value') == condition['value']
        )
    if condition['type'] == 'object_equals':
        return (
            condition['object_id'] is None and
            condition['property_name'] not in instance
        ) or (
            condition['object_id'] is not None and
            condition['property_name'] in instance and
            isinstance(instance[condition['property_name']], dict) and
            instance[condition['property_name']].get('object_id') == condition['object_id']
        )
    if condition['type'] == 'any':
        for sub_condition in condition['conditions']:
            if is_condition_fulfilled(sub_condition, instance):
                return True
        # any without passing conditions is considered False
        return False
    if condition['type'] == 'all':
        for sub_condition in condition['conditions']:
            if not is_condition_fulfilled(sub_condition, instance):
                return False
        # all without failing conditions is considered True
        return True
    if condition['type'] == 'not':
        return not is_condition_fulfilled(condition['condition'], instance)

    # unknown or unfulfillable condition
    return False


def are_conditions_fulfilled(
        conditions: typing.Optional[typing.List[typing.Dict[str, typing.Any]]],
        instance: typing.Dict[str, typing.Any]
) -> bool:
    """
    Check whether a list of conditions are fulfilled by the parent object.

    :param conditions: the list of conditions to check, or None
    :param instance: the instance to check the conditions on
    :return: whether or not all condition are fulfilled
    """
    return conditions is None or all(
        is_condition_fulfilled(condition, instance)
        for condition in conditions
    )



# coding: utf-8
"""
Add enable_instruments to location_types table.
"""

import flask_sqlalchemy

from .utils import table_has_column
from ..locations import LocationType


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('location_types', 'enable_instruments'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE location_types
            ADD enable_instruments BOOLEAN NOT NULL DEFAULT false
    """))
    db.session.execute(db.text("""
        UPDATE location_types
            SET enable_instruments = true
            WHERE id = :default_type_id
    """), params={'default_type_id': LocationType.LOCATION})
    db.session.execute(db.text("""
        ALTER TABLE location_types
            ALTER COLUMN enable_instruments DROP DEFAULT
    """))
    return True



# coding: utf-8
"""

"""

import enum

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'

import typing


class Permissions(enum.Enum):
    NONE = 0
    READ = 1
    WRITE = 2  # includes READ
    GRANT = 3  # includes READ and WRITE

    def __contains__(self, item: 'Permissions') -> bool:
        return self.value >= item.value

    def __str__(self) -> str:
        return self.name.lower()

    def __le__(self, other: 'Permissions') -> bool:
        if isinstance(other, Permissions):
            return self.value <= other.value
        return NotImplemented

    def __lt__(self, other: 'Permissions') -> bool:
        if isinstance(other, Permissions):
            return self.value < other.value
        return NotImplemented

    @staticmethod
    def from_name(
            name: str
    ) -> 'Permissions':
        members = {
            'none': Permissions.NONE,
            'read': Permissions.READ,
            'write': Permissions.WRITE,
            'grant': Permissions.GRANT,
        }
        try:
            return members[name.lower()]
        except KeyError:
            raise ValueError('Invalid name')

    @staticmethod
    def from_value(
            value: typing.Optional[int]
    ) -> 'Permissions':
        if value is None:
            return Permissions.NONE
        for member in Permissions:
            if member.value == value:
                return member
        return Permissions.NONE



# coding: utf-8
"""

"""

import enum
import datetime
import typing

from sqlalchemy.orm import Mapped, Query

from .. import db
from .files import File
from .utils import Model

if typing.TYPE_CHECKING:
    from ..logic.users import User

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


@enum.unique
class FileLogEntryType(enum.Enum):
    OTHER = 0
    EDIT_TITLE = 1
    EDIT_DESCRIPTION = 2
    HIDE_FILE = 3
    UNHIDE_FILE = 4
    EDIT_URL = 5


class FileLogEntry(Model):
    __tablename__ = 'file_log_entries'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    type: Mapped[FileLogEntryType] = db.Column(db.Enum(FileLogEntryType), nullable=False)
    object_id: Mapped[int] = db.Column(db.Integer, nullable=False)
    file_id: Mapped[int] = db.Column(db.Integer, nullable=False)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    data: Mapped[typing.Dict[str, typing.Any]] = db.Column(db.JSON, nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["FileLogEntry"]]

    __table_args__ = (
        db.ForeignKeyConstraint([object_id, file_id], [File.object_id, File.id]),
    )

    def __init__(
            self,
            object_id: int,
            file_id: int,
            user_id: int,
            type: FileLogEntryType,
            data: typing.Dict[str, typing.Any],
            utc_datetime: typing.Optional[datetime.datetime] = None
    ) -> None:
        super().__init__(
            type=type,
            object_id=object_id,
            file_id=file_id,
            user_id=user_id,
            data=data,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc)
        )
        self.user: typing.Optional['User'] = None

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, type={self.type}, file_id={self.file_id}, user_id={self.user_id}, utc_datetime={self.utc_datetime}, data={self.data})>'



# coding: utf-8
"""
Logic module for user managed groups

Users can create uniquely named groups and use them to more easily manage
permissions for samples or measurements. At this time, all group members
are equally able to add new members to a group and to modify a group's
name and description.

This module provides a procedural interface to the underlying database model,
with the immutable Group class accessible to the developer instead of an
SQLAlchemy ORM class.

As the group models use flask-sqlalchemy however, the functions in this
module should be called from within a Flask application context.
"""
import dataclasses
import datetime
import typing

import flask

from .. import db
from ..models import groups
from .users import check_user_exists, get_mutable_user
from .security_tokens import generate_token
from .notifications import create_notification_for_being_invited_to_a_group
from ..logic.languages import get_language_by_lang_code, Language
from . import errors


# group names are limited to this (semi-arbitrary) length
MAX_GROUP_NAME_LENGTH = 100


@dataclasses.dataclass(frozen=True)
class Group:
    """
    This class provides an immutable wrapper around models.groups.Group.
    """
    id: int
    name: typing.Dict[str, str]
    description: typing.Dict[str, str]

    @classmethod
    def from_database(cls, group: groups.Group) -> 'Group':
        return Group(
            id=group.id,
            name=group.name,
            description=group.description
        )


@dataclasses.dataclass(frozen=True)
class GroupInvitation:
    """
    This class provides an immutable wrapper around models.groups.GroupInvitation.
    """
    id: int
    group_id: int
    user_id: int
    inviter_id: int
    utc_datetime: datetime.datetime
    accepted: bool
    revoked: bool

    @classmethod
    def from_database(cls, group_invitation: groups.GroupInvitation) -> 'GroupInvitation':
        return GroupInvitation(
            id=group_invitation.id,
            group_id=group_invitation.group_id,
            user_id=group_invitation.user_id,
            inviter_id=group_invitation.inviter_id,
            utc_datetime=group_invitation.utc_datetime,
            accepted=group_invitation.accepted,
            revoked=group_invitation.revoked
        )

    @property
    def expired(self) -> bool:
        expiration_datetime = self.utc_datetime + datetime.timedelta(seconds=flask.current_app.config['INVITATION_TIME_LIMIT'])
        return bool(datetime.datetime.now(datetime.timezone.utc) >= expiration_datetime)


def create_group(
        name: typing.Union[str, typing.Dict[str, str]],
        description: typing.Union[str, typing.Dict[str, str]],
        initial_user_id: int
) -> Group:
    """
    Creates a new group with the given names and descriptions and adds an
    initial user to it.

    :param name: a dict containing unique names of the group. The keys are language codes of existing languages and the
        values are the names for the corresponding languages.
    :param description: a dict (possibly empty) containing descriptions for the group. The keys are language codes and
        the values are descriptions
    :param initial_user_id: the user ID of the initial group member
    :return: the newly created group
    :raise errors.InvalidGroupNameError: when the group name is empty or more
        than MAX_GROUP_NAME_LENGTH characters long
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    :raise errors.GroupAlreadyExistsError: when another group with the given
        name already exists
    :raise errors.LanguageDoesNotExistError: when there is no language for the given language code
    :raise errors.MissingEnglishTranslationError: when the english translation is missing
    """
    if isinstance(name, str):
        name = {
            'en': name
        }
    if isinstance(description, str):
        description = {
            'en': description
        }

    try:
        for language_code, name_text in list(name.items()):
            # check language code
            language = get_language_by_lang_code(language_code)
            # Check name
            if not 1 <= len(name_text) <= MAX_GROUP_NAME_LENGTH:
                if language.id != Language.ENGLISH and not name_text:
                    del name[language_code]
                else:
                    raise errors.InvalidGroupNameError()

            existing_group = groups.Group.query.filter(groups.Group.name[language_code].astext.cast(db.Unicode) == name_text).first()
            if existing_group is not None:
                raise errors.GroupAlreadyExistsError()
    except errors.LanguageDoesNotExistError:
        raise errors.LanguageDoesNotExistError("There is no language for the given lang code")
    except errors.GroupAlreadyExistsError:
        raise errors.GroupAlreadyExistsError()
    except errors.InvalidGroupNameError:
        raise errors.InvalidGroupNameError()
    if 'en' not in name:
        raise errors.MissingEnglishTranslationError()

    for item in list(description.items()):
        # delete empty descriptions
        if item[1] == '':
            del description[item[0]]

    user = get_mutable_user(initial_user_id)
    group = groups.Group(name=name, description=description)
    db.session.add(group)
    group.members.append(user)
    db.session.add(group)
    db.session.commit()
    return Group.from_database(group)


def update_group(
        group_id: int,
        name: typing.Dict[str, str],
        description: typing.Dict[str, str]
) -> None:
    """
    Updates the group's names and descriptions.

    :param group_id: the ID of an existing group
    :param name: the new dict containing unique group names. Keys are lang codes and values are names
    :param description: the new group descriptions in a dict. Keys are lang codes and values are descriptions
    :raise errors.InvalidGroupNameError: when the group name is empty or more
        than MAX_GROUP_NAME_LENGTH characters long
    :raise errors.GroupDoesNotExistError: when no group with the given
        group ID exists
    :raise errors.GroupAlreadyExistsError: when another group with the given
        name already exists
    :raise errors.LanguageDoesNotExistError: when there is no language with the given language code
    :raise errors.MissingEnglishTranslationError: when the english translation is missing
    """
    try:
        for language_code, name_text in list(name.items()):
            # check language code
            language = get_language_by_lang_code(language_code)
            # Check name
            if not 1 <= len(name_text) <= MAX_GROUP_NAME_LENGTH:
                if language.id != Language.ENGLISH and not name_text:
                    del name[language_code]
                else:
                    raise errors.InvalidGroupNameError()
            existing_group = groups.Group.query.filter(
                groups.Group.name[language_code].astext.cast(db.Unicode) == name_text
            ).first()
            if existing_group is not None and existing_group.id != group_id:
                raise errors.GroupAlreadyExistsError()
    except errors.LanguageDoesNotExistError:
        raise errors.LanguageDoesNotExistError("There is no language for the given lang code")
    except errors.GroupAlreadyExistsError:
        raise errors.GroupAlreadyExistsError()
    except errors.InvalidGroupNameError:
        raise errors.InvalidGroupNameError()
    if 'en' not in name:
        raise errors.MissingEnglishTranslationError()

    group = groups.Group.query.filter_by(id=group_id).first()
    if group is None:
        raise errors.GroupDoesNotExistError()

    for item in list(description.items()):
        # deleted empty descriptions
        if item[1] == '':
            del description[item[0]]

    group.name = name
    group.description = description
    db.session.add(group)
    db.session.commit()


def delete_group(group_id: int) -> None:
    """
    Deletes the given group, freeing up the name and removing all permissions
    that might have been granted to the group.

    This is also implicitly done when the last member leaves a group.

    :param group_id: the ID of an existing group
    :raise errors.GroupDoesNotExistError: when no group with the given
        group ID exists
    """
    group = groups.Group.query.filter_by(id=group_id).first()
    if group is None:
        raise errors.GroupDoesNotExistError()
    # group permissions and group default permissions will be deleted due to
    # ondelete = "CASCADE" in the model. No need to delete them manually here.
    db.session.delete(group)
    db.session.commit()


def get_group(group_id: int) -> Group:
    """
    Returns the group with the given ID.

    :param group_id: the ID of an existing group
    :return: the group
    :raise errors.GroupDoesNotExistError: when no group with the given
        group ID exists
    """
    group = groups.Group.query.filter_by(id=group_id).first()
    if group is None:
        raise errors.GroupDoesNotExistError()
    return Group.from_database(group)


def get_groups() -> typing.List[Group]:
    """
    Returns a list of all existing groups.

    :return: the list of groups
    """
    return [Group.from_database(group) for group in groups.Group.query.all()]


def get_group_member_ids(group_id: int) -> typing.List[int]:
    """
    Returns a list of the user IDs of all members of the group with the given
    group ID.

    :param group_id: the ID of an existing group
    :return: the member ID list
    :raise errors.GroupDoesNotExistError: when no group with the given
        group ID exists
    """
    group = groups.Group.query.filter_by(id=group_id).first()
    if group is None:
        raise errors.GroupDoesNotExistError()
    return [user.id for user in group.members]


def get_user_groups(user_id: int) -> typing.List[Group]:
    """
    Returns a list of the group IDs of all groups the user with the given
    user ID is a member of.

    :param user_id: the ID of an existing user
    :return: the member ID list
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """
    user = get_mutable_user(user_id)
    return [Group.from_database(group) for group in user.groups]


def invite_user_to_group(group_id: int, user_id: int, inviter_id: int) -> None:
    """
    Invite a user to a group.

    :param group_id: the ID of an existing group
    :param user_id: the ID of an existing user
    :param inviter_id: the ID of who invited this user to the group
    :raise errors.GroupDoesNotExistError: when no group with the given group
        ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        or inviter ID exists
    :raise errors.UserAlreadyMemberOfGroupError: when the user with the given
        user ID already is a member of this group
    """
    # ensure the inviter exists
    check_user_exists(inviter_id)
    group = groups.Group.query.filter_by(id=group_id).first()
    if group is None:
        raise errors.GroupDoesNotExistError()
    user = get_mutable_user(user_id)
    if user in group.members:
        raise errors.UserAlreadyMemberOfGroupError()

    invitation = groups.GroupInvitation(
        group_id=group_id,
        user_id=user_id,
        inviter_id=inviter_id,
        utc_datetime=datetime.datetime.now(datetime.timezone.utc)
    )
    db.session.add(invitation)
    db.session.commit()
    token = generate_token(
        {
            'invitation_id': invitation.id,
            'user_id': user.id,
            'group_id': group_id
        },
        salt='invite_to_group',
        secret_key=flask.current_app.config['SECRET_KEY']
    )
    expiration_time_limit = flask.current_app.config['INVITATION_TIME_LIMIT']
    expiration_utc_datetime = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(seconds=expiration_time_limit)
    confirmation_url = flask.url_for("frontend.group", group_id=group_id, token=token, _external=True)
    create_notification_for_being_invited_to_a_group(user_id, group_id, inviter_id, confirmation_url, expiration_utc_datetime, invitation_id=invitation.id)


def add_user_to_group(group_id: int, user_id: int) -> None:
    """
    Adds the user with the given user ID to the group with the given group ID.

    :param group_id: the ID of an existing group
    :param user_id: the ID of an existing user
    :raise errors.GroupDoesNotExistError: when no group with the given
        group ID exists
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    :raise errors.UserAlreadyMemberOfGroupError: when the user is already
        a member of the group
    """
    group = groups.Group.query.filter_by(id=group_id).first()
    if group is None:
        raise errors.GroupDoesNotExistError()
    user = get_mutable_user(user_id)
    if user in group.members:
        raise errors.UserAlreadyMemberOfGroupError()
    group.members.append(user)
    invitations = get_group_invitations(
        group_id=group_id,
        user_id=user_id,
        include_expired_invitations=False,
        include_accepted_invitations=False
    )
    mutable_invitations = groups.GroupInvitation.query.filter(
        groups.GroupInvitation.id.in_(
            [
                invitation.id
                for invitation in invitations
            ]
        )
    ).all()
    for mutable_invitation in mutable_invitations:
        mutable_invitation.accepted = True
        db.session.add(mutable_invitation)
    db.session.commit()


def remove_user_from_group(group_id: int, user_id: int) -> None:
    """
    Removes the user with the given user ID from the group with the given
    group ID.

    If there are no members left in the group, it is deleted and its name is
    freed up to be used again for a new group.

    :param group_id: the ID of an existing group
    :param user_id: the ID of an existing user
    :raise errors.GroupDoesNotExistError: when no group with the given
        group ID exists
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    :raise errors.UserNotMemberOfGroupError: when the user is not a member of
        the group
    """
    group = groups.Group.query.filter_by(id=group_id).first()
    if group is None:
        raise errors.GroupDoesNotExistError()
    user = get_mutable_user(user_id)
    if user not in group.members:
        raise errors.UserNotMemberOfGroupError()
    group.members.remove(user)
    if not group.members:
        db.session.delete(group)
    db.session.commit()


def get_group_invitations(
        group_id: int,
        user_id: typing.Optional[int] = None,
        include_accepted_invitations: bool = True,
        include_expired_invitations: bool = True,
        include_revoked_invitations: bool = True
) -> typing.List[GroupInvitation]:
    """
    Get all (current) invitations for a user to join a group.

    :param group_id: the ID of an existing group
    :param user_id: the ID of an existing user or None
    :param include_accepted_invitations: whether to include invitations that
        have already been accepted by the user
    :param include_expired_invitations: whether to include invitations that
        have already expired
    :param include_revoked_invitations: whether to include invitations that
        have been revoked
    :return: the list of group invitations
    :raise errors.GroupDoesNotExistError: when no group with the given
        group ID exists
    """
    group_invitations_query = groups.GroupInvitation.query.filter_by(group_id=group_id)
    if user_id is not None:
        group_invitations_query = group_invitations_query.filter_by(user_id=user_id)
    if not include_accepted_invitations:
        group_invitations_query = group_invitations_query.filter_by(accepted=False)
    if not include_revoked_invitations:
        group_invitations_query = group_invitations_query.filter_by(revoked=False)
    if not include_expired_invitations:
        expiration_time_limit = flask.current_app.config['INVITATION_TIME_LIMIT']
        expired_invitation_datetime = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(seconds=expiration_time_limit)
        group_invitations_query = group_invitations_query.filter(groups.GroupInvitation.utc_datetime > expired_invitation_datetime)
    group_invitations = group_invitations_query.order_by(groups.GroupInvitation.utc_datetime).all()
    if not group_invitations:
        # ensure that a group with this ID exists
        get_group(group_id)
    return [
        GroupInvitation.from_database(group_invitation)
        for group_invitation in group_invitations
    ]


def get_group_invitation(invitation_id: int) -> GroupInvitation:
    """
    Get an existing invitation.

    :param invitation_id: the ID of an existing invitation
    :return: the invitation
    :raise errors.GroupInvitationDoesNotExistError: when no invitation with
        the given ID exists
    """
    invitation = groups.GroupInvitation.query.filter_by(id=invitation_id).first()
    if invitation is None:
        raise errors.GroupInvitationDoesNotExistError()
    return GroupInvitation.from_database(invitation)


def revoke_group_invitation(invitation_id: int) -> None:
    """
    Mark a group invitation as revoked.

    :param invitation_id: the ID of an existing invitation
    :raise errors.GroupInvitationDoesNotExistError: when no invitation with
        the given ID exists
    """
    invitation = groups.GroupInvitation.query.filter_by(id=invitation_id).first()
    if invitation is None:
        raise errors.GroupInvitationDoesNotExistError()
    invitation.revoked = True
    db.session.add(invitation)
    db.session.commit()



import enum
import datetime
import typing

from sqlalchemy.orm import Mapped, Query, relationship
from sqlalchemy.dialects import postgresql

from .. import db
from .utils import Model

if typing.TYPE_CHECKING:
    from .users import User


@enum.unique
class UpdatableObjectsCheckStatus(enum.Enum):
    POSTED = 0
    IN_PROGRESS = 1
    DONE = 2


class UpdatableObjectsCheck(Model):
    __tablename__ = 'updatable_objects_checks'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    status: Mapped[UpdatableObjectsCheckStatus] = db.Column(db.Enum(UpdatableObjectsCheckStatus), nullable=False)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    user: Mapped['User'] = relationship('User')
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)
    action_ids: Mapped[typing.Optional[typing.List[int]]] = db.Column(postgresql.JSONB, nullable=True)
    result: Mapped[typing.Dict[str, typing.Any]] = db.Column(postgresql.JSONB, nullable=True)
    automatic_schema_update: Mapped['AutomaticSchemaUpdate'] = relationship('AutomaticSchemaUpdate', back_populates='updatable_objects_check')

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query['UpdatableObjectsCheck']]


@enum.unique
class AutomaticSchemaUpdateStatus(enum.Enum):
    POSTED = 0
    IN_PROGRESS = 1
    DONE = 2


class AutomaticSchemaUpdate(Model):
    __tablename__ = 'automatic_schema_updates'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    status: Mapped[AutomaticSchemaUpdateStatus] = db.Column(db.Enum(AutomaticSchemaUpdateStatus), nullable=False)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    user: Mapped['User'] = relationship('User')
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)
    updatable_objects_check_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('updatable_objects_checks.id'), nullable=False, unique=True)
    updatable_objects_check: Mapped[UpdatableObjectsCheck] = relationship('UpdatableObjectsCheck', back_populates='automatic_schema_update')
    object_ids: Mapped[typing.List[int]] = db.Column(postgresql.JSONB, nullable=False)
    result: Mapped[typing.Dict[str, typing.Any]] = db.Column(postgresql.JSONB, nullable=True)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query['AutomaticSchemaUpdate']]



import typing

import flask

from . import core
from .. import errors
from ... import logic
from ...models import ObjectLogEntry


def post_trigger_object_log_webhooks(
    object_log_entry: ObjectLogEntry
) -> None:
    if flask.current_app.config["ENABLE_BACKGROUND_TASKS"]:
        core.post_background_task(
            type='trigger_object_log_webhooks',
            data={'object_log_entry_id': object_log_entry.id, 'object_id': object_log_entry.object_id},
            auto_delete=True
        )
    else:
        handle_trigger_object_log_webhooks({'object_log_entry_id': object_log_entry.id, 'object_id': object_log_entry.object_id}, None)


def post_webhook_send(
    data: typing.Dict[str, typing.Any],
    webhook_id: int
) -> None:
    if flask.current_app.config["ENABLE_BACKGROUND_TASKS"]:
        core.post_background_task(
            type='webhook_send',
            data={'webhook_id': webhook_id, 'data': data},
            auto_delete=True
        )
    else:
        handle_webhook_send({'webhook_id': webhook_id, 'data': data}, None)


def handle_trigger_object_log_webhooks(
    data: typing.Dict[str, typing.Any],
    task_id: typing.Optional[int]
) -> typing.Tuple[bool, typing.Optional[dict[str, typing.Any]]]:
    webhooks = logic.webhooks.get_object_log_webhooks_for_object(data['object_id'])
    for webhook in webhooks:
        user = logic.users.get_user(webhook.user_id)
        if flask.current_app.config['ENABLE_WEBHOOKS_FOR_USERS'] or user.is_admin:
            object_log_entry = logic.object_log.get_object_log_entry(data['object_log_entry_id'], webhook.user_id)
            object_log_entry_json = logic.object_log.object_log_entry_to_json(object_log_entry)
            post_webhook_send(object_log_entry_json, webhook.id)

    return True, {}


def handle_webhook_send(
    data: typing.Dict[str, typing.Any],
    task_id: typing.Optional[int]
) -> typing.Tuple[bool, typing.Optional[dict[str, typing.Any]]]:
    try:
        webhook = logic.webhooks.get_webhook(data['webhook_id'])
        webhook.send_data(data['data'])
    except errors.WebhookConnectionException:
        return False, {}
    return True, {}



# coding: utf-8
"""
Add short_description column to topics table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('topics', 'short_description'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE topics
        ADD short_description JSON NOT NULL DEFAULT '{}'::json
    """))
    db.session.execute(db.text("""
        ALTER TABLE topics
        ALTER COLUMN short_description DROP DEFAULT
    """))
    return True



# coding: utf-8
"""

"""

import datetime
import typing

from sqlalchemy.orm import Mapped

from .. import db
from .utils import Model


class KeyPair(Model):
    __tablename__ = 'key_pairs'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    sk_bytes: Mapped[bytes] = db.Column(db.LargeBinary, nullable=False)
    pk_bytes: Mapped[bytes] = db.Column(db.LargeBinary, nullable=False)
    utc_datetime_created: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)

    def __init__(
            self,
            sk_bytes: bytes,
            pk_bytes: bytes,
            utc_datetime_created: typing.Optional[datetime.datetime] = None
    ):
        super().__init__(
            sk_bytes=sk_bytes,
            pk_bytes=pk_bytes,
            utc_datetime_created=utc_datetime_created if utc_datetime_created is not None else
            datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0)
        )



# coding: utf-8
"""

"""
import json
import typing

import flask
import flask_login
import requests
from flask_babel import _

from ...logic.federation.update import update_poke_component
from .. import frontend
from ... import logic
from ...logic import user_log
from ...logic.actions import get_action
from ...logic.caching import cache_per_request
from ...logic.components import get_component_by_uuid
from ...logic.errors import ObjectDoesNotExistError
from ...logic.object_permissions import get_user_object_permissions, get_object_permissions_for_all_users, get_object_permissions_for_anonymous_users, get_object_permissions_for_users, get_object_permissions_for_groups, get_object_permissions_for_projects, request_object_permissions, get_object_if_user_has_permissions
from ...logic.shares import get_shares_for_object, add_object_share, update_object_share
from ...logic.users import get_users, get_users_for_component
from ...logic.groups import get_group
from ...logic.objects import get_object, get_fed_object
from ...logic.projects import get_project
from ...logic.components import get_component, get_components
from .forms import CopyPermissionsForm, ObjectNewShareAccessForm, ObjectEditShareAccessForm
from ..permission_forms import PermissionsForm, UserPermissionsForm, GroupPermissionsForm, ProjectPermissionsForm, handle_permission_forms, set_up_permissions_forms
from ...utils import object_permissions_required, FlaskResponseT
from ..utils import get_user_if_exists, check_current_user_is_not_readonly, get_groups_form_data
from ...models import Permissions, Object


def on_unauthorized(object_id: int) -> FlaskResponseT:
    permissions_by_user = get_object_permissions_for_users(object_id)
    has_grant_user = any(
        Permissions.GRANT in permissions
        for permissions in permissions_by_user.values()
    )
    return flask.render_template('objects/unauthorized.html', object_id=object_id, has_grant_user=has_grant_user), 403


@cache_per_request()
def get_object_if_current_user_has_read_permissions(object_id: int, component_uuid: typing.Optional[str] = None) -> typing.Optional[Object]:
    return get_object_if_user_has_permissions(user_id=flask_login.current_user.id, permissions=Permissions.READ, object_id=object_id, component_uuid=component_uuid)


@cache_per_request()
def get_fed_object_if_current_user_has_read_permissions(fed_object_id: int, component_uuid: str) -> typing.Optional[Object]:
    user_id = flask_login.current_user.id
    component = get_component_by_uuid(component_uuid)
    try:
        fed_object = get_fed_object(fed_object_id, component_id=component.id)
        permissions = get_user_object_permissions(fed_object.id, user_id)
    except ObjectDoesNotExistError:
        return None
    if Permissions.READ not in permissions:
        return None
    return fed_object


@frontend.route('/objects/<int:object_id>/permissions/request', methods=['POST'])
@flask_login.login_required
def object_permissions_request(object_id: int) -> FlaskResponseT:
    current_permissions = get_user_object_permissions(object_id=object_id, user_id=flask_login.current_user.id)
    if Permissions.READ in current_permissions:
        flask.flash(_('You already have permissions to access this object.'), 'error')
        return flask.redirect(flask.url_for('.object', object_id=object_id))
    request_object_permissions(flask_login.current_user.id, object_id)
    flask.flash(_('Your request for permissions has been sent.'), 'success')
    return flask.redirect(flask.url_for('.objects'))


@frontend.route('/objects/<int:object_id>/permissions')
@object_permissions_required(Permissions.READ, on_unauthorized=on_unauthorized)
def object_permissions(object_id: int) -> FlaskResponseT:
    check_current_user_is_not_readonly()
    object = get_object(object_id)
    components = get_components()
    if object.action_id is not None:
        action = get_action(object.action_id)
        instrument = action.instrument
    else:
        action = None
        instrument = None
    user_permissions = get_object_permissions_for_users(object_id=object_id, include_instrument_responsible_users=False, include_groups=False, include_projects=False, include_readonly=False, include_admin_permissions=False)
    group_permissions = get_object_permissions_for_groups(object_id=object_id, include_projects=False)
    project_permissions = get_object_permissions_for_projects(object_id=object_id)
    all_user_permissions = get_object_permissions_for_all_users(object_id=object_id)
    anonymous_user_permissions = get_object_permissions_for_anonymous_users(object_id=object_id)
    component_policies = {share.component_id: share for share in get_shares_for_object(object_id)}
    policies = {share.component_id: share.policy for share in get_shares_for_object(object_id)}
    suggested_user_id_str = flask.request.args.get('add_user_id', '')
    try:
        suggested_user_id = int(suggested_user_id_str)
    except ValueError:
        suggested_user_id = None
    if Permissions.GRANT in get_user_object_permissions(object_id=object_id, user_id=flask_login.current_user.id):
        (
            add_user_permissions_form,
            add_group_permissions_form,
            add_project_permissions_form,
            permissions_form
        ) = set_up_permissions_forms(
            resource_permissions=logic.object_permissions.object_permissions,
            resource_id=object_id,
            existing_all_user_permissions=all_user_permissions,
            existing_anonymous_user_permissions=anonymous_user_permissions,
            existing_user_permissions=user_permissions,
            existing_group_permissions=group_permissions,
            existing_project_permissions=project_permissions
        )

        users = get_users(exclude_hidden=not flask_login.current_user.is_admin or not flask_login.current_user.settings['SHOW_HIDDEN_USERS_AS_ADMIN'], exclude_fed=True, exclude_eln_import=True)
        users = [user for user in users if user.id not in user_permissions]
        users.sort(key=lambda user: user.id)

        show_groups_form, groups_treepicker_info = get_groups_form_data(
            basic_group_filter=lambda group: group.id not in group_permissions
        )

        show_projects_form, projects_treepicker_info = get_groups_form_data(
            project_group_filter=lambda group: group.id not in project_permissions
        )

        possible_new_components = [component for component in components if component.id not in component_policies.keys()]
        component_users = {
            component.id: {
                user.fed_id: user.get_name(include_ref=True, use_local_identity=True)
                for user in get_users_for_component(component.id, exclude_hidden=False)
            }
            for component in components
        }
        add_component_policy_form = ObjectNewShareAccessForm()
        add_component_policy_form.data.data = True
        add_component_policy_form.action.data = True
        add_component_policy_form.users.data = True
        add_component_policy_form.files.data = True
        add_component_policy_form.comments.data = True
        add_component_policy_form.object_location_assignments.data = True
        edit_component_policy_form = ObjectEditShareAccessForm()
        edit_component_policy_form.data.data = True
        edit_component_policy_form.action.data = True
        edit_component_policy_form.users.data = True
        edit_component_policy_form.files.data = True
        edit_component_policy_form.comments.data = True
        edit_component_policy_form.object_location_assignments.data = True
        copy_permissions_form: typing.Optional[CopyPermissionsForm] = CopyPermissionsForm()
        assert copy_permissions_form is not None
        copy_permissions_form.object_id.choices = []
    else:
        permissions_form = None
        add_user_permissions_form = None
        add_group_permissions_form = None
        add_project_permissions_form = None
        add_component_policy_form = None
        copy_permissions_form = None
        edit_component_policy_form = None

        users = []
        possible_new_components = None
        component_users = {}
        show_groups_form = False
        groups_treepicker_info = None
        show_projects_form = False
        projects_treepicker_info = None

    return flask.render_template(
        'objects/object_permissions.html',
        instrument=instrument,
        action=action,
        object=object,
        user_permissions=user_permissions,
        group_permissions=group_permissions,
        project_permissions=project_permissions,
        all_user_permissions=all_user_permissions,
        anonymous_user_permissions=anonymous_user_permissions,
        federation_shares=component_policies,
        get_user=get_user_if_exists,
        get_component=get_component,
        Permissions=Permissions,
        permissions_form=permissions_form,
        users=users,
        show_groups_form=show_groups_form,
        groups_treepicker_info=groups_treepicker_info,
        show_projects_form=show_projects_form,
        projects_treepicker_info=projects_treepicker_info,
        edit_component_policy_form=edit_component_policy_form,
        component_users=component_users,
        copy_permissions_form=copy_permissions_form,
        add_component_policy_form=add_component_policy_form,
        possible_new_components=possible_new_components,
        add_user_permissions_form=add_user_permissions_form,
        add_group_permissions_form=add_group_permissions_form,
        get_group=get_group,
        add_project_permissions_form=add_project_permissions_form,
        get_project=get_project,
        suggested_user_id=suggested_user_id,
        json_dumps=json.dumps,
        policies=policies
    )


@frontend.route('/objects/<int:object_id>/permissions', methods=['POST'])
@object_permissions_required(Permissions.GRANT)
def update_object_permissions(object_id: int) -> FlaskResponseT:
    permissions_form = PermissionsForm()
    add_user_permissions_form = UserPermissionsForm()
    add_group_permissions_form = GroupPermissionsForm()
    add_project_permissions_form = ProjectPermissionsForm()
    add_component_policy_form = ObjectNewShareAccessForm()
    edit_component_policy_form = ObjectEditShareAccessForm()
    copy_permissions_form = CopyPermissionsForm()
    if 'copy_permissions' in flask.request.form:
        copy_permissions_form.object_id.choices = []
        if copy_permissions_form.validate_on_submit():
            logic.object_permissions.copy_permissions(object_id, int(copy_permissions_form.object_id.data))
            logic.object_permissions.set_user_object_permissions(object_id, flask_login.current_user.id, Permissions.GRANT)
            flask.flash(_("Successfully copied object permissions."), 'success')
    elif 'add_component_policy' in flask.request.form and add_component_policy_form.validate_on_submit():
        component_id = add_component_policy_form.component_id.data
        component = get_component(component_id)
        policy = {
            'access': {
                'data': add_component_policy_form.data.data,
                'action': add_component_policy_form.action.data,
                'users': add_component_policy_form.users.data,
                'files': add_component_policy_form.files.data,
                'comments': add_component_policy_form.comments.data,
                'object_location_assignments': add_component_policy_form.object_location_assignments.data,
            },
            'permissions': {'users': {}, 'groups': {}, 'projects': {}, 'all_users': 'none'}
        }
        allowed_permissions = [Permissions.READ.name.lower(), Permissions.WRITE.name.lower(), Permissions.GRANT.name.lower()]
        policy_permissions_are_valid = False
        for attr_name, value in flask.request.form.items():
            if attr_name.startswith('permissions_add_policy_user_'):
                user_id_str = attr_name[len('permissions_add_policy_user_'):]
                try:
                    user_id = int(user_id_str)
                except ValueError:
                    break
                if value not in allowed_permissions:
                    break
                policy['permissions']['users'][user_id] = value
            if attr_name.startswith('permissions_add_policy_group_'):
                basic_group_id_str = attr_name[len('permissions_add_policy_group_'):]
                try:
                    basic_group_id = int(basic_group_id_str)
                except ValueError:
                    break
                if value not in allowed_permissions:
                    break
                policy['permissions']['groups'][basic_group_id] = value
            if attr_name.startswith('permissions_add_policy_project_'):
                project_group_id_str = attr_name[len('permissions_add_policy_project_'):]
                try:
                    project_group_id = int(project_group_id_str)
                except ValueError:
                    break
                if value not in allowed_permissions:
                    break
                policy['permissions']['projects'][project_group_id] = value
        else:
            policy_permissions_are_valid = True
        if not policy_permissions_are_valid:
            flask.flash(_("A problem occurred while changing the object permissions. Please try again."), 'error')
            return flask.redirect(flask.url_for('.object_permissions', object_id=object_id))
        add_object_share(object_id, component_id, policy, user_id=flask_login.current_user.id)
        try:
            update_poke_component(component)
        except logic.errors.MissingComponentAddressError:
            flask.flash(_('Unable to contact %(component_name)s. Missing database address.', component_name=component.get_name()), 'warning')
        except logic.errors.NoAuthenticationMethodError:
            flask.flash(_('No valid authentication method configured for %(component_name)s (%(component_address)s).', component_name=component.get_name(), component_address=component.address), 'warning')
        except requests.ConnectionError:
            flask.flash(_('Unable to contact %(component_name)s (%(component_address)s).', component_name=component.get_name(), component_address=component.address), 'warning')
        flask.flash(_("Successfully updated object permissions."), 'success')
    elif 'edit_component_policy' in flask.request.form and edit_component_policy_form.validate_on_submit():
        component_id = edit_component_policy_form.component_id.data
        component = get_component(component_id)
        policy = {
            'access': {
                'data': edit_component_policy_form.data.data,
                'action': edit_component_policy_form.action.data,
                'users': edit_component_policy_form.users.data,
                'files': edit_component_policy_form.files.data,
                'comments': edit_component_policy_form.comments.data,
                'object_location_assignments': edit_component_policy_form.object_location_assignments.data,
            },
            'permissions': {'users': {}, 'groups': {}, 'projects': {}, 'all_users': 'none'}
        }
        allowed_permissions = [Permissions.READ.name.lower(), Permissions.WRITE.name.lower(), Permissions.GRANT.name.lower()]
        policy_permissions_are_valid = False
        for attr_name, value in flask.request.form.items():
            if attr_name.startswith('permissions_edit_policy_user_'):
                user_id_str = attr_name[len('permissions_edit_policy_user_'):]
                try:
                    user_id = int(user_id_str)
                except ValueError:
                    break
                if value not in allowed_permissions:
                    break
                policy['permissions']['users'][user_id] = value
            if attr_name.startswith('permissions_edit_policy_group_'):
                basic_group_id_str = attr_name[len('permissions_edit_policy_group_'):]
                try:
                    basic_group_id = int(basic_group_id_str)
                except ValueError:
                    break
                if value not in allowed_permissions:
                    break
                policy['permissions']['groups'][basic_group_id] = value
            if attr_name.startswith('permissions_edit_policy_project_'):
                project_group_id_str = attr_name[len('permissions_edit_policy_project_'):]
                try:
                    project_group_id = int(project_group_id_str)
                except ValueError:
                    break
                if value not in allowed_permissions:
                    break
                policy['permissions']['projects'][project_group_id] = value
        else:
            policy_permissions_are_valid = True
        if not policy_permissions_are_valid:
            flask.flash(_("A problem occurred while changing the object permissions. Please try again."), 'error')
            return flask.redirect(flask.url_for('.object_permissions', object_id=object_id))
        update_object_share(object_id, component_id, policy, user_id=flask_login.current_user.id)
        try:
            update_poke_component(component)
        except logic.errors.MissingComponentAddressError:
            flask.flash(_('Unable to contact %(component_name)s. Missing database address.', component_name=component.get_name()), 'warning')
        except logic.errors.NoAuthenticationMethodError:
            flask.flash(_('No valid authentication method configured for %(component_name)s (%(component_address)s).', component_name=component.get_name(), component_address=component.address), 'warning')
        except requests.ConnectionError:
            flask.flash(_('Unable to contact %(component_name)s (%(component_address)s).', component_name=component.get_name(), component_address=component.address), 'warning')
        flask.flash(_("Successfully updated object permissions."), 'success')
    else:
        if handle_permission_forms(
            logic.object_permissions.object_permissions,
            object_id,
            add_user_permissions_form,
            add_group_permissions_form,
            add_project_permissions_form,
            permissions_form
        ):
            user_log.edit_object_permissions(user_id=flask_login.current_user.id, object_id=object_id)
            flask.flash(_("Successfully updated object permissions."), 'success')
        else:
            flask.flash(_("A problem occurred while changing the object permissions. Please try again."), 'error')
    return flask.redirect(flask.url_for('.object_permissions', object_id=object_id))



import typing
from http import HTTPStatus

import flask
import flask_login
import wtforms
from flask_babel import _
from flask_wtf import FlaskForm
from wtforms import StringField, SelectMultipleField
from wtforms.fields.numeric import IntegerField
from wtforms.validators import InputRequired, AnyOf, ValidationError

from .. import frontend
from ...utils import FlaskResponseT
from ...logic import errors
from ...logic.topics import get_topics
from ...logic.action_permissions import get_sorted_actions_for_user
from ...logic.automatic_schema_updates import get_updatable_objects_checks, get_updatable_objects_check, start_updatable_objects_checks, get_automatic_schema_updates, start_automatic_schema_updates
from ...logic.objects import get_max_object_id
from ...logic.object_permissions import get_objects_with_permissions
from ...models import UpdatableObjectsCheckStatus, AutomaticSchemaUpdateStatus, Permissions


class CheckForUpdatableObjectsForm(FlaskForm):
    action_ids = SelectMultipleField()
    submit = StringField(validators=[InputRequired(), AnyOf(['check_for_updatable_objects_form'])])


class StartSchemaUpdatesForm(FlaskForm):
    updatable_objects_check_id = IntegerField(validators=[InputRequired()])
    object_ids = SelectMultipleField()
    submit = StringField(validators=[InputRequired(), AnyOf(['start_schema_updates_form'])])

    def validate_updatable_objects_check_id(self, field: wtforms.IntegerField) -> None:
        updatable_objects_check_id = field.data
        try:
            get_updatable_objects_check(updatable_objects_check_id=updatable_objects_check_id)
        except errors.UpdatableObjectsCheckDoesNotExistError:
            raise ValidationError('Invalid updatable_objects_check_id')


@frontend.route('/admin/automatic_schema_updates/', methods=['GET', 'POST'])
@flask_login.login_required
def automatic_schema_updates() -> FlaskResponseT:
    if not flask_login.current_user.is_admin:
        return flask.abort(HTTPStatus.FORBIDDEN)
    updatable_objects_checks = get_updatable_objects_checks()
    automatic_schema_updates = get_automatic_schema_updates()

    checking_in_progress = any(
        updatable_objects_check.status != UpdatableObjectsCheckStatus.DONE
        for updatable_objects_check in updatable_objects_checks
    )
    update_in_progress = any(
        automatic_schema_update.status != AutomaticSchemaUpdateStatus.DONE
        for automatic_schema_update in automatic_schema_updates
    )

    all_actions = None
    sorted_action_topics = None
    if update_in_progress or checking_in_progress or flask_login.current_user.is_readonly:
        check_for_updatable_objects_form = None
    else:
        check_for_updatable_objects_form = CheckForUpdatableObjectsForm()

        all_actions = get_sorted_actions_for_user(
            user_id=flask_login.current_user.id
        )
        check_for_updatable_objects_form.action_ids.choices = [
            (str(action.id), str(action.id))
            for action in all_actions
        ]
        sorted_action_topics = []
        if not flask.current_app.config['DISABLE_TOPICS']:
            sorted_topics = get_topics()
            for topic in sorted_topics:
                for action in all_actions:
                    if topic in action.topics:
                        sorted_action_topics.append(topic)
                        break

        if check_for_updatable_objects_form.validate_on_submit():
            try:
                filter_action_ids: typing.Optional[typing.List[int]] = [
                    int(action_id)
                    for action_id in check_for_updatable_objects_form.action_ids.data
                ]
            except ValueError:
                flask.flash(_('Invalid action ID.'), 'error')
                return flask.redirect(flask.url_for('.automatic_schema_updates'))
            if not filter_action_ids:
                filter_action_ids = None
            try:
                start_updatable_objects_checks(flask_login.current_user.id, filter_action_ids)
                flask.flash(_('An updatable objects check has been started. Please refresh this page in a few minutes.'), 'success')
            except errors.UpdatableObjectsCheckAlreadyInProgressError:
                flask.flash(_('Another updatable objects check is already in progress.'), 'error')
            return flask.redirect(flask.url_for('.automatic_schema_updates'))

    start_schema_updates_form = None
    updatable_objects_by_id = {}
    if not checking_in_progress and not update_in_progress and updatable_objects_checks:
        latest_updatable_objects_check = updatable_objects_checks[0]
        if latest_updatable_objects_check.result and not latest_updatable_objects_check.automatic_schema_update:
            updatable_object_ids = {
                object_id
                for object_id, version_id, messages in latest_updatable_objects_check.result.get('automatically_updatable_objects', [])
            }.union({
                object_id
                for object_id, version_id, messages in latest_updatable_objects_check.result.get('manually_updatable_objects', [])
            })
            updatable_objects = get_objects_with_permissions(
                user_id=flask_login.current_user.id,
                permissions=Permissions.GRANT,
                object_ids=list(updatable_object_ids),
                name_only=True
            )
            updatable_objects_by_id = {
                updatable_object.object_id: updatable_object
                for updatable_object in updatable_objects
            }
            start_schema_updates_form = StartSchemaUpdatesForm()
            start_schema_updates_form.object_ids.choices = [
                (str(object_id), str(object_id))
                for object_id in updatable_objects_by_id
            ]
            if start_schema_updates_form.validate_on_submit():
                object_ids_to_update = {
                    int(object_id_str)
                    for object_id_str in start_schema_updates_form.object_ids.data
                }
                start_automatic_schema_updates(
                    updatable_objects_check_id=start_schema_updates_form.updatable_objects_check_id.data,
                    user_id=flask_login.current_user.id,
                    object_ids=list(object_ids_to_update)
                )
                flask.flash(_('An automatic schema update has been started. Please refresh this page in a few minutes.'), 'success')
                return flask.redirect(flask.url_for('.automatic_schema_updates'))
            if updatable_object_ids != set(updatable_objects_by_id.keys()) and not flask_login.current_user.has_admin_permissions:
                flask.flash(_('Without enabling the setting "Use Admin Permissions" in your preferences, you will not be able to update all updatable objects.'), 'warning')

    max_object_id = get_max_object_id()
    return flask.render_template(
        'admin/automatic_schema_updates.html',
        checking_in_progress=checking_in_progress,
        update_in_progress=update_in_progress,
        check_for_updatable_objects_form=check_for_updatable_objects_form,
        updatable_object_checks=updatable_objects_checks,
        max_object_id=max_object_id,
        UpdatableObjectsCheckStatus=UpdatableObjectsCheckStatus,
        updatable_objects_by_id=updatable_objects_by_id,
        start_schema_updates_form=start_schema_updates_form,
        automatic_schema_updates=automatic_schema_updates,
        AutomaticSchemaUpdateStatus=AutomaticSchemaUpdateStatus,
        all_actions=all_actions,
        sorted_action_topics=sorted_action_topics,
        current_status={
            'updatable_objects_checks': {
                updatable_objects_check.id: updatable_objects_check.status == UpdatableObjectsCheckStatus.DONE
                for updatable_objects_check in updatable_objects_checks
            },
            'automatic_schema_updates': {
                automatic_schema_update.id: automatic_schema_update.status == AutomaticSchemaUpdateStatus.DONE
                for automatic_schema_update in automatic_schema_updates
            }
        }
    )


@frontend.route('/admin/automatic_schema_updates/status', methods=['GET', 'POST'])
@flask_login.login_required
def automatic_schema_updates_status() -> FlaskResponseT:
    if not flask_login.current_user.is_admin:
        return flask.abort(HTTPStatus.FORBIDDEN)
    updatable_objects_checks = get_updatable_objects_checks()
    automatic_schema_updates = get_automatic_schema_updates()
    return flask.jsonify({
        'updatable_objects_checks': {
            updatable_objects_check.id: updatable_objects_check.status == UpdatableObjectsCheckStatus.DONE
            for updatable_objects_check in updatable_objects_checks
        },
        'automatic_schema_updates': {
            automatic_schema_update.id: automatic_schema_update.status == AutomaticSchemaUpdateStatus.DONE
            for automatic_schema_update in automatic_schema_updates
        }
    })




from http import HTTPStatus

import flask
import flask_login

from . import frontend
from ..logic.utils import show_numeric_tags_warning, do_numeric_tags_exist
from ..utils import FlaskResponseT


@frontend.route('/admin/warnings/')
@flask_login.login_required
def admin_warnings() -> FlaskResponseT:
    if not flask_login.current_user.is_admin:
        return flask.abort(HTTPStatus.FORBIDDEN)
    warnings = {}
    for deprecation_warning_check in [
        show_numeric_tags_warning,
    ]:
        warnings[deprecation_warning_check.__name__] = deprecation_warning_check()

    has_warnings = any(warnings.values())
    return flask.render_template(
        'admin/warnings.html',
        has_warnings=has_warnings,
        do_numeric_tags_exist=do_numeric_tags_exist() if show_numeric_tags_warning() else False,
        **warnings
    )



# coding: utf-8
"""
Add preview_image_binary_data and preview_image_mime_type columns to files table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('files', 'preview_image_binary_data'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE files
        ADD preview_image_binary_data BYTEA NULL,
        ADD preview_image_mime_type VARCHAR NULL
    """))
    return True
